#include "Camera.h"

using namespace Eigen;

const float PI = 3.14159265358979323846f;

//=================================================================================================
/**
*	Creates a view projection matrix for the provided camera properties.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
const Matrix4f& Camera::update(
	const Vector2f& viewPortSize,
	const Vector4f& position,
	const Vector4f& forward,
	const Vector4f& up,
	float verticalFovDeg,
	float aspect,
	float zNear,
	float zFar)
{
	mAspectRatio = aspect;
	mZNear = zNear;
	mZFar = zFar;
	mViewPortSize = viewPortSize;
	mPosition = position;
	mForward = forward;
	mRight.head<3>() = mForward.head<3>().cross(up.head<3>());
	mRight.z() = 0.0f;
	mRight.w() = 0.0f;
	mRight.normalize();
	mUp.head<3>() = mRight.head<3>().cross(mForward.head<3>()).normalized();

	mVerticalFieldOfView = verticalFovDeg * PI / 180.0f;
	mHorizontalFieldOfView = 2.0f * std::atan(std::tan(mVerticalFieldOfView * 0.5f) * mAspectRatio);

	mViewMatrix = createViewMatrix();
	mInverseViewMatrix = mViewMatrix.inverse();
	mProjectionMatrix = createPerspectiveProjection();
	mViewProjectionMatrix = mProjectionMatrix * mViewMatrix;

	return mViewProjectionMatrix;
}

//=================================================================================================
/**
*	Creates a view matrix to transform world space vertices into the camera space defined by the
*	provided parameters.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
Matrix4f Camera::createViewMatrix()
{
	Matrix4f m;
	m(0, 0) = mRight.x();		m(0, 1) = mRight.y();		m(0, 2) = mRight.z();		m(0, 3) = -mPosition.dot(mRight);
	m(1, 0) = mUp.x();			m(1, 1) = mUp.y();			m(1, 2) = mUp.z();			m(1, 3) = -mPosition.dot(mUp);
	m(2, 0) = -mForward.x();	m(2, 1) = -mForward.y();	m(2, 2) = -mForward.z();	m(2, 3) = -mPosition.dot(-mForward);
	m(3, 0) = 0.0f;				m(3, 1) = 0.0f;				m(3, 2) = 0.0f;				m(3, 3) = 1.0f;

	return std::move(m);
}

//=================================================================================================
/**
*	Creates a perspective projection matrix to transform view space vertices into the clip space.
*	Note: This projection is not a standard OpenGL projection matrix, but an inversed one, meaning
*	that it maps the far plane to 0 instead of 1. This allows us to drastically increase the depth
*	buffer precision (for a detailed explanation, see
*	http://outerra.blogspot.co.at/2012/11/maximizing-depth-buffer-range-and.html and
*	https://www.opengl.org/discussion_boards/showthread.php/
*	185674-How-to-setup-projectionMatrix-worldMatrix-for-use-with-inverted-float-depth-buffer).
*
*	@author	Daniel Cornel
**/
//=================================================================================================
Matrix4f Camera::createPerspectiveProjection()
{
	Matrix4f m = Matrix4f::Zero();
	m(0, 0) = 1.0f / (mAspectRatio * std::tan(0.5f * mVerticalFieldOfView));
	m(1, 1) = 1.0f / std::tan(0.5f * mVerticalFieldOfView);
	m(2, 2) = mZNear / (mZFar - mZNear);
	m(3, 2) = -1.0f;
	m(2, 3) = mZFar * mZNear / (mZFar - mZNear);

	return std::move(m);
}