#pragma once
#ifndef ShaderCodesCornelEtAl2022_h
#define ShaderCodesCornelEtAl2022_h

namespace ShaderCodesCornelEtAl2022
{
	#include "../CommonShaders.h"

	const char* CommonCornelEtAl2022 = MULTILINE(
		const int FLAG_SUBDIVIDE = 1 << 0;
		const int FLAG_MERGE = 1 << 1;
		const int FLAG_DELETED = 1 << 2;

		const u8vec2[] nextGraphColors = u8vec2[](
			u8vec2(4u, 6u),
			u8vec2(5u, 7u),
			u8vec2(7u, 4u),
			u8vec2(6u, 5u),
			u8vec2(2u, 0u),
			u8vec2(3u, 1u),
			u8vec2(0u, 3u),
			u8vec2(1u, 2u),
			u8vec2(0xFFu));

		u8vec2 getNextGraphColors(int parentGraphColor)
		{
			return nextGraphColors[min(parentGraphColor, 8)];
		}
	);

	const char* AlignedBox = MULTILINE(
		// This file is a simplified copy of the Eigen::AlignedBox2f

		struct AlignedBox2f
		{
			vec2 mMin;
			vec2 mMax;
		};

		vec2 diagonal(AlignedBox2f box)
		{
			return box.mMax - box.mMin;
		}

		bool isEmpty(AlignedBox2f box)
		{
			return any(greaterThan(box.mMin, box.mMax));
		}

		bool isApprox(AlignedBox2f box, AlignedBox2f otherBox, vec2 tolerance)
		{
			return all(lessThanEqual(abs(box.mMin - otherBox.mMin), tolerance)) && all(lessThanEqual(abs(box.mMax - otherBox.mMax), tolerance));
		}

		bool contains(AlignedBox2f box, vec2 p)
		{
			return all(lessThanEqual(box.mMin, p)) && all(lessThanEqual(p, box.mMax));
		}

		bool contains(AlignedBox2f box, AlignedBox2f otherBox)
		{
			return all(lessThanEqual(box.mMin, otherBox.mMin)) && all(lessThanEqual(otherBox.mMax, box.mMax));
		}

		bool intersects(AlignedBox2f box, AlignedBox2f otherBox)
		{
			return all(lessThanEqual(box.mMin, otherBox.mMax)) && all(lessThanEqual(otherBox.mMin, box.mMax));
		}

		AlignedBox2f intersection(AlignedBox2f box, AlignedBox2f otherBox)
		{
			AlignedBox2f newBox;
			newBox.mMin = max(box.mMin, otherBox.mMin);
			newBox.mMax = min(box.mMax, otherBox.mMax);

			return newBox;
		}

		AlignedBox2f extend(AlignedBox2f box, vec2 p)
		{
			AlignedBox2f newBox;
			newBox.mMin = min(box.mMin, p);
			newBox.mMax = max(box.mMax, p);

			return newBox;
		}

		AlignedBox2f extend(AlignedBox2f box, AlignedBox2f otherBox)
		{
			AlignedBox2f newBox;
			newBox.mMin = min(box.mMin, otherBox.mMin);
			newBox.mMax = max(box.mMax, otherBox.mMax);

			return newBox;
		}
	);

	const char* LoDFunction = MULTILINE(
		uniform float squaredMinimumEdgeLength;
		uniform float upperTriangleSize;
		uniform float lowerTriangleSize;

		float distanceSquared(vec4 a, vec4 b)
		{
			a -= b;

			return dot(a, a);
		}

		float getEdgeLengthInPixels(vec4 worldSpacePosition0, vec4 worldSpacePosition1)
		{
			float worldSpaceEdgeLength = distance(worldSpacePosition0, worldSpacePosition1);
			vec4 worldSpaceCenterPosition = 0.5 * (worldSpacePosition0 + worldSpacePosition0);

			vec4 viewSpaceCenterPosition = viewMatrix * worldSpaceCenterPosition;
			vec4 viewSpaceOffsetPosition = viewSpaceCenterPosition + vec4(worldSpaceEdgeLength, 0.0, 0.0, 0.0);

			vec4 clipSpaceCenterPosition = projectionMatrix * viewSpaceCenterPosition;
			vec4 clipSpaceOffsetPosition = projectionMatrix * viewSpaceOffsetPosition;

			vec2 pixelSpaceCenterPosition = (0.5 * clipSpaceCenterPosition.xy / clipSpaceCenterPosition.w + 0.5) * viewPortSize;
			vec2 pixelSpaceOffsetPosition = (0.5 * clipSpaceOffsetPosition.xy / clipSpaceOffsetPosition.w + 0.5) * viewPortSize;

			return distance(pixelSpaceCenterPosition, pixelSpaceOffsetPosition);
		}

		bool inFrustum(vec4 worldSpacePosition, float maxDistance)
		{
			return !((dot(worldSpacePosition, frustumPlaneLeft) <= maxDistance) ||
				(dot(worldSpacePosition, frustumPlaneRight) <= maxDistance) ||
				(dot(worldSpacePosition, frustumPlaneBottom) <= maxDistance) ||
				(dot(worldSpacePosition, frustumPlaneTop) <= maxDistance) ||
				(dot(worldSpacePosition, frustumPlaneFar) <= maxDistance) ||
				(dot(worldSpacePosition, frustumPlaneNear) <= maxDistance));
		}

		uint8_t getSubdivisionFlags(ivec3 vertices)
		{
			vec4 v0 = vec4(vertexBuffer_fetch(vertices.x), 1.0);
			vec4 v1 = vec4(vertexBuffer_fetch(vertices.y), 1.0);
			vec4 v2 = vec4(vertexBuffer_fetch(vertices.z), 1.0);

			vec2 worldSpaceTriangleAABBMin = min(v0.xy, min(v1.xy, v2.xy));
			vec2 worldSpaceTriangleAABBMax = max(v0.xy, max(v1.xy, v2.xy));
			vec2 worldSpaceTriangleExtents = worldSpaceTriangleAABBMax - worldSpaceTriangleAABBMin;
			float maxExtent = max(worldSpaceTriangleExtents.x, worldSpaceTriangleExtents.y);
			float maxDistance = -4.0 * maxExtent;

			if ((!inFrustum(v0, maxDistance)) && (!inFrustum(v1, maxDistance)) && (!inFrustum(v2, maxDistance)))
			{
				return uint8_t(FLAG_MERGE);
			}

			vec3 pixelSpaceEdgeLengths = vec3(
				getEdgeLengthInPixels(v0, v1),
				getEdgeLengthInPixels(v1, v2),
				getEdgeLengthInPixels(v2, v0));

			bvec3 greaterThanPixelSpaceUpper = greaterThan(pixelSpaceEdgeLengths, vec3(upperTriangleSize));
			bvec3 lessThanPixelSpaceLower = lessThan(pixelSpaceEdgeLengths, vec3(lowerTriangleSize));
			bvec3 greaterThanWorldSpaceMin = greaterThan(vec3(distanceSquared(v1, v0), distanceSquared(v2, v1), distanceSquared(v0, v2)), vec3(squaredMinimumEdgeLength));

			return (any((greaterThanWorldSpaceMin) && (greaterThanPixelSpaceUpper))) ?
				uint8_t(FLAG_SUBDIVIDE) :
				uint8_t(all(lessThanPixelSpaceLower) ? FLAG_MERGE : 0);
		}
	);

	const std::vector<const char*> InitialTessellationCompute
	{
		MULTILINE(
			#version 450\n
			#extension GL_NV_gpu_shader5 : enable\n

			layout(local_size_x = 32, local_size_y = 1, local_size_z = 1) in;

			layout(std430) restrict writeonly buffer vertexBufferBlock
			{
				float vertexBuffer[];
			};

			void vertexBuffer_store(int elementIndex, vec3 value)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				vertexBuffer[elementOffsets.x] = value.x;
				vertexBuffer[elementOffsets.y] = value.y;
				vertexBuffer[elementOffsets.z] = value.z;
			}

			layout(std430) restrict writeonly buffer subdivisionFlagsBufferBlock
			{
				uint8_t subdivisionFlagsBuffer[];
			};

			layout(std430) restrict writeonly buffer parentTriangleBufferBlock
			{
				int parentTriangleBuffer[];
			};

			layout(std430) restrict writeonly buffer firstChildTriangleBufferBlock
			{
				int firstChildTriangleBuffer[];
			};

			layout(std430) restrict writeonly buffer triangleGraphColorBufferBlock
			{
				uint8_t triangleGraphColorBuffer[];
			};

			layout(std430) restrict writeonly buffer adjacentTrianglesBufferBlock
			{
				int adjacentTrianglesBuffer[];
			};

			void adjacentTrianglesBuffer_store(int elementIndex, ivec3 value)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				adjacentTrianglesBuffer[elementOffsets.x] = value.x;
				adjacentTrianglesBuffer[elementOffsets.y] = value.y;
				adjacentTrianglesBuffer[elementOffsets.z] = value.z;
			}

			layout(std430) restrict writeonly buffer vertexIndicesBufferBlock
			{
				int vertexIndicesBuffer[];
			};

			void vertexIndicesBuffer_store(int elementIndex, ivec3 value)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				vertexIndicesBuffer[elementOffsets.x] = value.x;
				vertexIndicesBuffer[elementOffsets.y] = value.y;
				vertexIndicesBuffer[elementOffsets.z] = value.z;
			}

			uniform vec2 domainUpper;
			uniform vec2 chunkGridCellSize;
			uniform ivec2 chunkGridDimensions;
			uniform int numThreads;
		),
		AlignedBox,
		HeightfieldSampling,
		MULTILINE(
			void main()
			{
				int chunkCellIndex = int(gl_GlobalInvocationID.x);

				if (chunkCellIndex >= numThreads)
				{
					return;
				}

				int y = chunkCellIndex / chunkGridDimensions.x;
				int x = chunkCellIndex - y * chunkGridDimensions.x;

				int triangle0Index = (y * chunkGridDimensions.x + x) << 1;
				int triangle1Index = triangle0Index + 1;

				int bottomLeftVertexIndex = y * (chunkGridDimensions.x + 1) + x;
				int bottomRightVertexIndex = bottomLeftVertexIndex + 1;
				int topLeftVertexIndex = (y + 1) * (chunkGridDimensions.x + 1) + x;
				int topRightVertexIndex = topLeftVertexIndex + 1;

				AlignedBox2f offsetChunkQuad;
				offsetChunkQuad.mMin = vec2(x, y) * chunkGridCellSize + domainOrigin;
				offsetChunkQuad.mMax = offsetChunkQuad.mMin + chunkGridCellSize;

				AlignedBox2f chunkQuad;
				chunkQuad.mMin = offsetChunkQuad.mMin;
				chunkQuad.mMax = min(offsetChunkQuad.mMax, domainUpper);

				// Do the actual tessellation with two triangles per chunk grid cell

				vec2 chunkBottomRight = vec2(chunkQuad.mMax.x, chunkQuad.mMin.y);
				vec2 chunkTopLeft = vec2(chunkQuad.mMin.x, chunkQuad.mMax.y);

				vertexBuffer_store(bottomLeftVertexIndex, vec3(chunkQuad.mMin, sampleHeightfield(chunkQuad.mMin)));
				vertexBuffer_store(bottomRightVertexIndex, vec3(chunkBottomRight, sampleHeightfield(chunkBottomRight)));
				vertexBuffer_store(topLeftVertexIndex, vec3(chunkTopLeft, sampleHeightfield(chunkTopLeft)));
				vertexBuffer_store(topRightVertexIndex, vec3(chunkQuad.mMax, sampleHeightfield(chunkQuad.mMax)));

				subdivisionFlagsBuffer[triangle0Index] = uint8_t(0);
				subdivisionFlagsBuffer[triangle1Index] = uint8_t(0);

				parentTriangleBuffer[triangle0Index] = -1;
				parentTriangleBuffer[triangle1Index] = -1;

				firstChildTriangleBuffer[triangle0Index] = -1;
				firstChildTriangleBuffer[triangle1Index] = -1;

				bool flippedChunk = (x & 1) != (y & 1);

				triangleGraphColorBuffer[triangle0Index] = (flippedChunk) ? uint8_t(2) : uint8_t(0);
				triangleGraphColorBuffer[triangle1Index] = (flippedChunk) ? uint8_t(3) : uint8_t(1);

				adjacentTrianglesBuffer_store(
					triangle0Index,
					(flippedChunk) ?
					ivec3((x == 0) ? -1 : (triangle0Index - 1), triangle1Index, (y == chunkGridDimensions.y - 1) ? -1 : (((y + 1) * chunkGridDimensions.x + x) << 1)) :
					ivec3((y == 0) ? -1 : (((y - 1) * chunkGridDimensions.x + x) << 1), triangle1Index, (x == 0) ? -1 : (triangle0Index - 1)));
				adjacentTrianglesBuffer_store(
					triangle1Index,
					(flippedChunk) ?
					ivec3((x == chunkGridDimensions.x - 1) ? -1 : (triangle1Index + 1), triangle0Index, (y == 0) ? -1 : (((y - 1) * chunkGridDimensions.x + x) << 1) + 1) :
					ivec3((y == chunkGridDimensions.y - 1) ? -1 : ((((y + 1) * chunkGridDimensions.x + x) << 1) + 1), triangle0Index, (x == chunkGridDimensions.x - 1) ? -1 : (triangle1Index + 1)));

				vertexIndicesBuffer_store(
					triangle0Index,
					(flippedChunk) ?
					ivec3(topLeftVertexIndex, bottomLeftVertexIndex, topRightVertexIndex) :
					ivec3(bottomLeftVertexIndex, bottomRightVertexIndex, topLeftVertexIndex));
				vertexIndicesBuffer_store(
					triangle1Index,
					(flippedChunk) ?
					ivec3(bottomRightVertexIndex, topRightVertexIndex, bottomLeftVertexIndex) :
					ivec3(topRightVertexIndex, topLeftVertexIndex, bottomRightVertexIndex));
			}
		)
	};

	const std::vector<const char*> UpdateSubdivisionFlagsCompute
	{
		MULTILINE(
			#version 450\n
			#extension GL_NV_gpu_shader5 : enable\n

			layout(local_size_x = 32, local_size_y = 1, local_size_z = 1) in;

			layout(std430) restrict readonly buffer vertexBufferBlock
			{
				float vertexBuffer[];
			};

			vec3 vertexBuffer_fetch(int elementIndex)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				return vec3(vertexBuffer[elementOffsets.x], vertexBuffer[elementOffsets.y], vertexBuffer[elementOffsets.z]);
			}

			layout(std430) restrict buffer subdivisionFlagsBufferBlock
			{
				uint8_t subdivisionFlagsBuffer[];
			};

			layout(std430) restrict readonly buffer vertexIndicesBufferBlock
			{
				int vertexIndicesBuffer[];
			};

			ivec3 vertexIndicesBuffer_fetch(int elementIndex)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				return ivec3(vertexIndicesBuffer[elementOffsets.x], vertexIndicesBuffer[elementOffsets.y], vertexIndicesBuffer[elementOffsets.z]);
			}

			uniform int numThreads;
		),
		CommonCornelEtAl2022,
		CameraUniforms,
		AlignedBox,
		LoDFunction,
		MULTILINE(
			void main()
			{
				int triangleIndex = int(gl_GlobalInvocationID.x);

				if (triangleIndex >= numThreads)
				{
					return;
				}

				if ((int(subdivisionFlagsBuffer[triangleIndex]) & FLAG_DELETED) != 0)
				{
					return;
				}

				subdivisionFlagsBuffer[triangleIndex] = getSubdivisionFlags(vertexIndicesBuffer_fetch(triangleIndex));
			}
		)
	};

	const std::vector<const char*> CompactionCompute
	{
		MULTILINE(
			#version 450\n
			#extension GL_NV_gpu_shader5 : enable\n

			layout(local_size_x = 32, local_size_y = 1, local_size_z = 1) in;

			layout(std430) restrict readonly buffer vertexBufferBlock
			{
				float vertexBuffer[];
			};

			vec3 vertexBuffer_fetch(int elementIndex)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				return vec3(vertexBuffer[elementOffsets.x], vertexBuffer[elementOffsets.y], vertexBuffer[elementOffsets.z]);
			}

			layout(std430) restrict readonly buffer subdivisionFlagsBufferBlock
			{
				uint8_t subdivisionFlagsBuffer[];
			};

			layout(std430) restrict readonly buffer firstChildTriangleBufferBlock
			{
				int firstChildTriangleBuffer[];
			};

			layout(std430) restrict readonly buffer vertexIndicesBufferBlock
			{
				int vertexIndicesBuffer[];
			};

			ivec3 vertexIndicesBuffer_fetch(int elementIndex)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				return ivec3(vertexIndicesBuffer[elementOffsets.x], vertexIndicesBuffer[elementOffsets.y], vertexIndicesBuffer[elementOffsets.z]);
			}

			// This struct is defined by OpenGL, do not change it!
			struct DrawElementsIndirectCommand
			{
				uint primitiveCount;
				uint instanceCount;
				uint firstIndex;
				uint baseVertex;
				uint baseInstance;
			};

			layout(std430) restrict buffer indirectDrawBufferBlock
			{
				DrawElementsIndirectCommand indirectDrawBuffer;
			};

			layout(std430) restrict writeonly buffer visibleTriangleBufferBlock
			{
				int visibleTriangleBuffer[];
			};

			void visibleTriangleBuffer_store(int elementIndex, ivec3 value)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				visibleTriangleBuffer[elementOffsets.x] = value.x;
				visibleTriangleBuffer[elementOffsets.y] = value.y;
				visibleTriangleBuffer[elementOffsets.z] = value.z;
			}

			uniform int numThreads;
		),
		CommonCornelEtAl2022,
		CameraUniforms,
		MULTILINE(
			bool inFrustum(vec4 worldSpacePosition, float maxDistance)
			{
				return !((dot(worldSpacePosition, frustumPlaneLeft) <= maxDistance) ||
					(dot(worldSpacePosition, frustumPlaneRight) <= maxDistance) ||
					(dot(worldSpacePosition, frustumPlaneBottom) <= maxDistance) ||
					(dot(worldSpacePosition, frustumPlaneTop) <= maxDistance) ||
					(dot(worldSpacePosition, frustumPlaneFar) <= maxDistance) ||
					(dot(worldSpacePosition, frustumPlaneNear) <= maxDistance));
			}

			void main()
			{
				int triangleIndex = int(gl_GlobalInvocationID.x);

				if (triangleIndex >= numThreads)
				{
					return;
				}

				if ((int(subdivisionFlagsBuffer[triangleIndex]) & FLAG_DELETED) != 0)
				{
					return;
				}

				if (firstChildTriangleBuffer[triangleIndex] > -1)
				{
					return;
				}

				ivec3 vertexIndices = vertexIndicesBuffer_fetch(triangleIndex);

				if ((inFrustum(vec4(vertexBuffer_fetch(vertexIndices.x), 1.0), 0.0)) ||
					(inFrustum(vec4(vertexBuffer_fetch(vertexIndices.y), 1.0), 0.0)) ||
					(inFrustum(vec4(vertexBuffer_fetch(vertexIndices.z), 1.0), 0.0)))
				{
					uint nextFreeIndex = atomicAdd(indirectDrawBuffer.primitiveCount, 3u);

					visibleTriangleBuffer_store(int(nextFreeIndex) / 3, vertexIndices);
				}
			}
		)
	};

	const std::vector<const char*> SubdivisionTriangleSelectionCompute
	{
		MULTILINE(
			#version 450\n
			#extension GL_NV_gpu_shader5 : enable\n

			layout(local_size_x = 32, local_size_y = 1, local_size_z = 1) in;

			layout(std430) restrict readonly buffer subdivisionFlagsBufferBlock
			{
				uint8_t subdivisionFlagsBuffer[];
			};

			layout(std430) restrict readonly buffer firstChildTriangleBufferBlock
			{
				int firstChildTriangleBuffer[];
			};

			layout(std430) restrict readonly buffer adjacentTrianglesBufferBlock
			{
				int adjacentTrianglesBuffer[];
			};

			ivec3 adjacentTrianglesBuffer_fetch(int elementIndex)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				return ivec3(adjacentTrianglesBuffer[elementOffsets.x], adjacentTrianglesBuffer[elementOffsets.y], adjacentTrianglesBuffer[elementOffsets.z]);
			}

			int adjacentTrianglesBuffer_fetchY(int elementIndex)
			{
				return adjacentTrianglesBuffer[3 * elementIndex + 1];
			}

			layout(std430) restrict readonly buffer triangleGraphColorBufferBlock
			{
				uint8_t triangleGraphColorBuffer[];
			};

			layout(std430) restrict writeonly buffer graphColor0BufferBlock
			{
				int graphColor0Buffer[];
			};

			layout(std430) restrict buffer graphColor0CounterBlock
			{
				int graphColor0Counter;
			};

			layout(std430) restrict writeonly buffer graphColor1BufferBlock
			{
				int graphColor1Buffer[];
			};

			layout(std430) restrict buffer graphColor1CounterBlock
			{
				int graphColor1Counter;
			};

			layout(std430) restrict writeonly buffer graphColor2BufferBlock
			{
				int graphColor2Buffer[];
			};

			layout(std430) restrict buffer graphColor2CounterBlock
			{
				int graphColor2Counter;
			};

			layout(std430) restrict writeonly buffer graphColor3BufferBlock
			{
				int graphColor3Buffer[];
			};

			layout(std430) restrict buffer graphColor3CounterBlock
			{
				int graphColor3Counter;
			};

			uniform int numThreads;
		),
		CommonCornelEtAl2022,
		MULTILINE(
			void main()
			{
				int currentTriangleIndex = int(gl_GlobalInvocationID.x);

				if (currentTriangleIndex >= numThreads)
				{
					return;
				}

				int currentTriangleSubdivisionFlags = int(subdivisionFlagsBuffer[currentTriangleIndex]);

				// Ignore deleted triangles
				if ((currentTriangleSubdivisionFlags & FLAG_DELETED) != 0)
				{
					return;
				}

				// Ignore triangles with children
				if (firstChildTriangleBuffer[currentTriangleIndex] > -1)
				{
					return;
				}

				ivec3 currentTriangleAdjacentTriangles = adjacentTrianglesBuffer_fetch(currentTriangleIndex);
				int neighbor1TriangleIndex = currentTriangleAdjacentTriangles.y;
				bool hasNeighbor1 = (neighbor1TriangleIndex > -1);

				int neighbor1TriangleFirstChildTriangle = -1;
				ivec3 neighbor1TriangleAdjacentTriangles = ivec3(-1);
				int neighbor1TriangleSubdivisionFlags = 0;

				if (hasNeighbor1)
				{
					neighbor1TriangleFirstChildTriangle = firstChildTriangleBuffer[neighbor1TriangleIndex];

					// Ignore triangles with children
					if (neighbor1TriangleFirstChildTriangle > -1)
					{
						return;
					}

					neighbor1TriangleAdjacentTriangles = adjacentTrianglesBuffer_fetch(neighbor1TriangleIndex);

					// Only split two triangles that share the same hypothenuse
					if (neighbor1TriangleAdjacentTriangles.y != currentTriangleIndex)
					{
						return;
					}

					neighbor1TriangleSubdivisionFlags = int(subdivisionFlagsBuffer[neighbor1TriangleIndex]);
				}

				// First, check whether the hypothenuse needs to be split
				int subdivisionFlags = currentTriangleSubdivisionFlags | neighbor1TriangleSubdivisionFlags;

				// If not, check whether one of the other edges needs to be split
				// and is the hypothenuse of an adjacent triangle
				if ((subdivisionFlags & FLAG_SUBDIVIDE) == 0)
				{
					ivec4 neighborsNeighbor1Indices = ivec4(-1);

					if (currentTriangleAdjacentTriangles.x > -1)
					{
						neighborsNeighbor1Indices.x = adjacentTrianglesBuffer_fetchY(currentTriangleAdjacentTriangles.x);
					}

					if (currentTriangleAdjacentTriangles.z > -1)
					{
						neighborsNeighbor1Indices.y = adjacentTrianglesBuffer_fetchY(currentTriangleAdjacentTriangles.z);
					}

					if (neighbor1TriangleAdjacentTriangles.x > -1)
					{
						neighborsNeighbor1Indices.z = adjacentTrianglesBuffer_fetchY(neighbor1TriangleAdjacentTriangles.x);
					}

					if (neighbor1TriangleAdjacentTriangles.z > -1)
					{
						neighborsNeighbor1Indices.w = adjacentTrianglesBuffer_fetchY(neighbor1TriangleAdjacentTriangles.z);
					}

					bvec4 equalsCurrentTriangleIndex = equal(neighborsNeighbor1Indices, ivec4(currentTriangleIndex));

					if (equalsCurrentTriangleIndex.x)
					{
						subdivisionFlags |= int(subdivisionFlagsBuffer[currentTriangleAdjacentTriangles.x]) & FLAG_SUBDIVIDE;
					}

					if (equalsCurrentTriangleIndex.y)
					{
						subdivisionFlags |= int(subdivisionFlagsBuffer[currentTriangleAdjacentTriangles.z]) & FLAG_SUBDIVIDE;
					}

					if (equalsCurrentTriangleIndex.z)
					{
						subdivisionFlags |= int(subdivisionFlagsBuffer[neighbor1TriangleAdjacentTriangles.x]) & FLAG_SUBDIVIDE;
					}

					if (equalsCurrentTriangleIndex.w)
					{
						subdivisionFlags |= int(subdivisionFlagsBuffer[neighbor1TriangleAdjacentTriangles.z]) & FLAG_SUBDIVIDE;
					}
				}

				// No need for subdivision
				if ((subdivisionFlags & FLAG_SUBDIVIDE) == 0)
				{
					return;
				}

				// In general, we only want to process triangles of the current graph color, with one
				// exception of triangles at the borders of a non-watertight mesh, i.e., triangles that
				// do not have a neighbor sharing the hypothenuse. If the non-existent neighbor would
				// have had the current graph color, we need to process these current triangles also in
				// this pass.
				int triangleGraphColor0 = int(triangleGraphColorBuffer[currentTriangleIndex]);
				int graphColor = (hasNeighbor1) ? triangleGraphColor0 : (triangleGraphColor0 & ~1);

				if ((graphColor & 1) != 0)
				{
					return;
				}

				if (graphColor == 0)
				{
					graphColor0Buffer[atomicAdd(graphColor0Counter, 1)] = currentTriangleIndex;
				}
				else if (graphColor == 2)
				{
					graphColor1Buffer[atomicAdd(graphColor1Counter, 1)] = currentTriangleIndex;
				}
				else if (graphColor == 4)
				{
					graphColor2Buffer[atomicAdd(graphColor2Counter, 1)] = currentTriangleIndex;
				}
				else
				{
					graphColor3Buffer[atomicAdd(graphColor3Counter, 1)] = currentTriangleIndex;
				}
			}
		)
	};

	const std::vector<const char*> SubdivisionCompute
	{
		MULTILINE(
			#version 450\n
			#extension GL_NV_gpu_shader5 : enable\n

			layout(local_size_x = 32, local_size_y = 1, local_size_z = 1) in;

			layout(std430) restrict buffer vertexBufferBlock
			{
				float vertexBuffer[];
			};

			vec3 vertexBuffer_fetch(int elementIndex)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				return vec3(vertexBuffer[elementOffsets.x], vertexBuffer[elementOffsets.y], vertexBuffer[elementOffsets.z]);
			}

			void vertexBuffer_store(int elementIndex, vec3 value)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				vertexBuffer[elementOffsets.x] = value.x;
				vertexBuffer[elementOffsets.y] = value.y;
				vertexBuffer[elementOffsets.z] = value.z;
			}

			layout(std430) restrict buffer subdivisionFlagsBufferBlock
			{
				uint8_t subdivisionFlagsBuffer[];
			};

			layout(std430) restrict buffer firstChildTriangleBufferBlock
			{
				int firstChildTriangleBuffer[];
			};

			layout(std430) restrict buffer parentTriangleBufferBlock
			{
				int parentTriangleBuffer[];
			};

			layout(std430) restrict buffer adjacentTrianglesBufferBlock
			{
				int adjacentTrianglesBuffer[];
			};

			ivec3 adjacentTrianglesBuffer_fetch(int elementIndex)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				return ivec3(adjacentTrianglesBuffer[elementOffsets.x], adjacentTrianglesBuffer[elementOffsets.y], adjacentTrianglesBuffer[elementOffsets.z]);
			}

			void adjacentTrianglesBuffer_store(int elementIndex, ivec3 value)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				adjacentTrianglesBuffer[elementOffsets.x] = value.x;
				adjacentTrianglesBuffer[elementOffsets.y] = value.y;
				adjacentTrianglesBuffer[elementOffsets.z] = value.z;
			}

			void adjacentTrianglesBuffer_storeComponent(int elementIndex, int componentIndex, int value)
			{
				adjacentTrianglesBuffer[3 * elementIndex + componentIndex] = value;
			}

			layout(std430) restrict buffer triangleGraphColorBufferBlock
			{
				uint8_t triangleGraphColorBuffer[];
			};

			layout(std430) restrict buffer vertexIndicesBufferBlock
			{
				int vertexIndicesBuffer[];
			};

			ivec3 vertexIndicesBuffer_fetch(int elementIndex)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				return ivec3(vertexIndicesBuffer[elementOffsets.x], vertexIndicesBuffer[elementOffsets.y], vertexIndicesBuffer[elementOffsets.z]);
			}

			void vertexIndicesBuffer_store(int elementIndex, ivec3 value)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				vertexIndicesBuffer[elementOffsets.x] = value.x;
				vertexIndicesBuffer[elementOffsets.y] = value.y;
				vertexIndicesBuffer[elementOffsets.z] = value.z;
			}

			layout(std430) restrict readonly buffer freeTriangleIndicesBufferBlock
			{
				int freeTriangleIndicesBuffer[];
			};

			layout(std430) restrict readonly buffer freeVertexIndicesBufferBlock
			{
				int freeVertexIndicesBuffer[];
			};

			layout(std430) restrict readonly buffer graphColorBufferBlock
			{
				int graphColorBuffer[];
			};

			uniform int triangleGraphColorNumInterleaved;
			uniform int triangleGraphColorInterleavedOffset;
			uniform int triangleGraphColorEndOffset;
			uniform int vertexGraphColorNumInterleaved;
			uniform int vertexGraphColorInterleavedOffset;
			uniform int vertexGraphColorEndOffset;
			uniform int numThreads;
		),
		CommonCornelEtAl2022,
		HeightfieldSampling,
		CameraUniforms,
		AlignedBox,
		LoDFunction,
		MULTILINE(
			void main()
			{
				int threadIndex = int(gl_GlobalInvocationID.x);

				if (threadIndex >= numThreads)
				{
					return;
				}

				int currentTriangleIndex = graphColorBuffer[threadIndex];
				ivec3 currentTriangleAdjacentTriangles = adjacentTrianglesBuffer_fetch(currentTriangleIndex);

				int neighbor1TriangleIndex = currentTriangleAdjacentTriangles.y;
				bool hasNeighbor1 = (neighbor1TriangleIndex > -1);
				ivec3 neighbor1TriangleAdjacentTriangles = ivec3(-1);

				if (hasNeighbor1)
				{
					neighbor1TriangleAdjacentTriangles = adjacentTrianglesBuffer_fetch(neighbor1TriangleIndex);
				}

				int newTriangleIndex;

				if (threadIndex < triangleGraphColorNumInterleaved)
				{
					newTriangleIndex = freeTriangleIndicesBuffer[triangleGraphColorInterleavedOffset - threadIndex];
				}
				else
				{
					newTriangleIndex = triangleGraphColorEndOffset + ((threadIndex - triangleGraphColorNumInterleaved) << 2);
				}

				int newVertexIndex;

				if (threadIndex < vertexGraphColorNumInterleaved)
				{
					newVertexIndex = freeVertexIndicesBuffer[vertexGraphColorInterleavedOffset - threadIndex];
				}
				else
				{
					newVertexIndex = vertexGraphColorEndOffset + threadIndex - vertexGraphColorNumInterleaved;
				}

				ivec3 currentTriangleVertexIndices = vertexIndicesBuffer_fetch(currentTriangleIndex);
				ivec3 neighbor1TriangleVertexIndices = ivec3(0);

				if (hasNeighbor1)
				{
					neighbor1TriangleVertexIndices = vertexIndicesBuffer_fetch(neighbor1TriangleIndex);
				}

				vec3 triangle0Vertex1 = vertexBuffer_fetch(currentTriangleVertexIndices.y);
				vec3 triangle0Vertex2 = vertexBuffer_fetch(currentTriangleVertexIndices.z);

				vec3 intersectionVertex;
				intersectionVertex.xy = 0.5 * (triangle0Vertex1.xy + triangle0Vertex2.xy);

				float height = sampleHeightfield(intersectionVertex.xy);
				intersectionVertex.z = height;

				vertexBuffer_store(newVertexIndex, intersectionVertex);

				ivec3 newTriangleIndices = ivec3(newTriangleIndex) + ivec3(1, 2, 3);

				ivec3 triangle0Child0AdjacentTriangles = ivec3((hasNeighbor1) ? newTriangleIndices.z : -1, currentTriangleAdjacentTriangles.z, newTriangleIndices.x);
				ivec3 triangle0Child0VertexIndices = ivec3(newVertexIndex, currentTriangleVertexIndices.zx);
				uint8_t triangle0Child0SubdivisionFlags = getSubdivisionFlags(triangle0Child0VertexIndices);

				ivec3 triangle0Child1AdjacentTriangles = ivec3(newTriangleIndex, currentTriangleAdjacentTriangles.x, (hasNeighbor1) ? newTriangleIndices.y : -1);
				ivec3 triangle0Child1VertexIndices = ivec3(newVertexIndex, currentTriangleVertexIndices.xy);
				uint8_t triangle0Child1SubdivisionFlags = getSubdivisionFlags(triangle0Child1VertexIndices);

				ivec3 triangle1Child0AdjacentTriangles = ivec3(newTriangleIndices.x, neighbor1TriangleAdjacentTriangles.z, newTriangleIndices.z);
				ivec3 triangle1Child0VertexIndices = ivec3(newVertexIndex, neighbor1TriangleVertexIndices.zx);
				uint8_t triangle1Child0SubdivisionFlags = (neighbor1TriangleIndex == -1) ? uint8_t(FLAG_DELETED) : getSubdivisionFlags(triangle1Child0VertexIndices);

				ivec3 triangle1Child1AdjacentTriangles = ivec3(newTriangleIndices.y, neighbor1TriangleAdjacentTriangles.x, newTriangleIndex);
				ivec3 triangle1Child1VertexIndices = ivec3(newVertexIndex, neighbor1TriangleVertexIndices.xy);
				uint8_t triangle1Child1SubdivisionFlags = (neighbor1TriangleIndex == -1) ? uint8_t(FLAG_DELETED) : getSubdivisionFlags(triangle1Child1VertexIndices);

				int triangleGraphColor1 = 0xFF;

				if (hasNeighbor1)
				{
					triangleGraphColor1 = int(triangleGraphColorBuffer[neighbor1TriangleIndex]);
				}

				int triangleGraphColor0 = int(triangleGraphColorBuffer[currentTriangleIndex]);
				u8vec2 triangle0ChildrenGraphColors = getNextGraphColors(triangleGraphColor0);
				u8vec2 triangle1ChildrenGraphColors = getNextGraphColors(triangleGraphColor1);

				triangleGraphColorBuffer[newTriangleIndex] = triangle0ChildrenGraphColors.x;
				triangleGraphColorBuffer[newTriangleIndices.x] = triangle0ChildrenGraphColors.y;
				triangleGraphColorBuffer[newTriangleIndices.y] = triangle1ChildrenGraphColors.x;
				triangleGraphColorBuffer[newTriangleIndices.z] = triangle1ChildrenGraphColors.y;

				parentTriangleBuffer[newTriangleIndex] = currentTriangleIndex;
				parentTriangleBuffer[newTriangleIndices.x] = currentTriangleIndex;
				parentTriangleBuffer[newTriangleIndices.y] = neighbor1TriangleIndex;
				parentTriangleBuffer[newTriangleIndices.z] = neighbor1TriangleIndex;

				adjacentTrianglesBuffer_store(newTriangleIndex, triangle0Child0AdjacentTriangles);
				adjacentTrianglesBuffer_store(newTriangleIndices.x, triangle0Child1AdjacentTriangles);
				adjacentTrianglesBuffer_store(newTriangleIndices.y, triangle1Child0AdjacentTriangles);
				adjacentTrianglesBuffer_store(newTriangleIndices.z, triangle1Child1AdjacentTriangles);

				ivec3 neighbor0TriangleAdjacentTriangles = ivec3(-1);
				ivec3 neighbor2TriangleAdjacentTriangles = ivec3(-1);
				ivec3 triangle1Neighbor0TriangleAdjacentTriangles = ivec3(-1);
				ivec3 triangle1Neighbor2TriangleAdjacentTriangles = ivec3(-1);

				vertexIndicesBuffer_store(newTriangleIndex, triangle0Child0VertexIndices);
				vertexIndicesBuffer_store(newTriangleIndices.x, triangle0Child1VertexIndices);
				vertexIndicesBuffer_store(newTriangleIndices.y, triangle1Child0VertexIndices);
				vertexIndicesBuffer_store(newTriangleIndices.z, triangle1Child1VertexIndices);

				subdivisionFlagsBuffer[newTriangleIndex] = triangle0Child0SubdivisionFlags;
				subdivisionFlagsBuffer[newTriangleIndices.x] = triangle0Child1SubdivisionFlags;
				subdivisionFlagsBuffer[newTriangleIndices.y] = triangle1Child0SubdivisionFlags;
				subdivisionFlagsBuffer[newTriangleIndices.z] = triangle1Child1SubdivisionFlags;

				firstChildTriangleBuffer[newTriangleIndex] = -1;
				firstChildTriangleBuffer[newTriangleIndices.x] = -1;
				firstChildTriangleBuffer[newTriangleIndices.y] = -1;
				firstChildTriangleBuffer[newTriangleIndices.z] = -1;

				firstChildTriangleBuffer[currentTriangleIndex] = newTriangleIndex;

				if (hasNeighbor1)
				{
					firstChildTriangleBuffer[neighbor1TriangleIndex] = newTriangleIndices.y;
				}

				if (currentTriangleAdjacentTriangles.x > -1)
				{
					neighbor0TriangleAdjacentTriangles = adjacentTrianglesBuffer_fetch(currentTriangleAdjacentTriangles.x);
				}

				if (currentTriangleAdjacentTriangles.z > -1)
				{
					neighbor2TriangleAdjacentTriangles = adjacentTrianglesBuffer_fetch(currentTriangleAdjacentTriangles.z);
				}

				if (hasNeighbor1)
				{
					if (neighbor1TriangleAdjacentTriangles.x > -1)
					{
						triangle1Neighbor0TriangleAdjacentTriangles = adjacentTrianglesBuffer_fetch(neighbor1TriangleAdjacentTriangles.x);
					}

					if (neighbor1TriangleAdjacentTriangles.z > -1)
					{
						triangle1Neighbor2TriangleAdjacentTriangles = adjacentTrianglesBuffer_fetch(neighbor1TriangleAdjacentTriangles.z);
					}
				}

				for (int i = 0; i < 3; i++)
				{
					if (neighbor0TriangleAdjacentTriangles[i] == currentTriangleIndex)
					{
						adjacentTrianglesBuffer_storeComponent(currentTriangleAdjacentTriangles.x, i, newTriangleIndices.x);
					}

					if (neighbor2TriangleAdjacentTriangles[i] == currentTriangleIndex)
					{
						adjacentTrianglesBuffer_storeComponent(currentTriangleAdjacentTriangles.z, i, newTriangleIndex);
					}

					if (hasNeighbor1)
					{
						if (triangle1Neighbor0TriangleAdjacentTriangles[i] == neighbor1TriangleIndex)
						{
							adjacentTrianglesBuffer_storeComponent(neighbor1TriangleAdjacentTriangles.x, i, newTriangleIndices.z);
						}

						if (triangle1Neighbor2TriangleAdjacentTriangles[i] == neighbor1TriangleIndex)
						{
							adjacentTrianglesBuffer_storeComponent(neighbor1TriangleAdjacentTriangles.z, i, newTriangleIndices.y);
						}
					}
				}
			}
		)
	};

	const std::vector<const char*> MergeTriangleSelectionCompute
	{
		MULTILINE(
			#version 450\n
			#extension GL_NV_gpu_shader5 : enable\n

			layout(local_size_x = 32, local_size_y = 1, local_size_z = 1) in;

			layout(std430) restrict readonly buffer subdivisionFlagsBufferBlock
			{
				uint8_t subdivisionFlagsBuffer[];
			};

			layout(std430) restrict readonly buffer firstChildTriangleBufferBlock
			{
				int firstChildTriangleBuffer[];
			};

			layout(std430) restrict readonly buffer parentTriangleBufferBlock
			{
				int parentTriangleBuffer[];
			};

			layout(std430) restrict readonly buffer adjacentTrianglesBufferBlock
			{
				int adjacentTrianglesBuffer[];
			};

			ivec3 adjacentTrianglesBuffer_fetch(int elementIndex)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				return ivec3(adjacentTrianglesBuffer[elementOffsets.x], adjacentTrianglesBuffer[elementOffsets.y], adjacentTrianglesBuffer[elementOffsets.z]);
			}

			layout(std430) restrict readonly buffer triangleGraphColorBufferBlock
			{
				uint8_t triangleGraphColorBuffer[];
			};

			layout(std430) restrict writeonly buffer graphColor0BufferBlock
			{
				int graphColor0Buffer[];
			};

			layout(std430) restrict buffer graphColor0CounterBlock
			{
				int graphColor0Counter;
			};

			layout(std430) restrict writeonly buffer graphColor1BufferBlock
			{
				int graphColor1Buffer[];
			};

			layout(std430) restrict buffer graphColor1CounterBlock
			{
				int graphColor1Counter;
			};

			layout(std430) restrict writeonly buffer graphColor2BufferBlock
			{
				int graphColor2Buffer[];
			};

			layout(std430) restrict buffer graphColor2CounterBlock
			{
				int graphColor2Counter;
			};

			layout(std430) restrict writeonly buffer graphColor3BufferBlock
			{
				int graphColor3Buffer[];
			};

			layout(std430) restrict buffer graphColor3CounterBlock
			{
				int graphColor3Counter;
			};

			uniform int numThreads;
		),
		CommonCornelEtAl2022,
		MULTILINE(
			void main()
			{
				int currentTriangleIndex = int(gl_GlobalInvocationID.x);

				if (currentTriangleIndex >= numThreads)
				{
					return;
				}

				int currentTriangleSubdivisionFlags = int(subdivisionFlagsBuffer[currentTriangleIndex]);

				// Skip deleted triangles
				if ((currentTriangleSubdivisionFlags & FLAG_DELETED) != 0)
				{
					return;
				}

				// Cannot merge triangles with children
				if (firstChildTriangleBuffer[currentTriangleIndex] > -1)
				{
					return;
				}

				int currentTriangleParentTriangle = parentTriangleBuffer[currentTriangleIndex];

				// Cannot merge root triangles
				if (currentTriangleParentTriangle == -1)
				{
					return;
				}

				int parent0TriangleIndex = currentTriangleParentTriangle;
				ivec3 parent0AdjacentTriangles = adjacentTrianglesBuffer_fetch(parent0TriangleIndex);

				int parent1TriangleIndex = parent0AdjacentTriangles[1];
				bool hasParent1 = (parent1TriangleIndex > -1);

				ivec3 parent1AdjacentTriangles = ivec3(-1);

				if (hasParent1)
				{
					parent1AdjacentTriangles = adjacentTrianglesBuffer_fetch(parent1TriangleIndex);
				}

				// Cannot merge triangles of which parents do not share a hypothenuse
				if ((hasParent1) && (parent1AdjacentTriangles[1] != parent0TriangleIndex))
				{
					return;
				}

				int parent0FirstChildTriangle = firstChildTriangleBuffer[parent0TriangleIndex];

				// Merging has to be done once for four triangles, so only do this operation in
				// the first triangle of the first parent triangle
				if (parent0FirstChildTriangle != currentTriangleIndex)
				{
					return;
				}

				int parent0SecondChildTriangle = parent0FirstChildTriangle + 1;
				int parent1FirstChildTriangle = -1;
				int parent1SecondChildTriangle = -1;

				if (hasParent1)
				{
					parent1FirstChildTriangle = firstChildTriangleBuffer[parent1TriangleIndex];
					parent1SecondChildTriangle = parent1FirstChildTriangle + 1;
				}

				ivec4 parentsFirstChildTriangles = ivec4(
					firstChildTriangleBuffer[parent0FirstChildTriangle],
					firstChildTriangleBuffer[parent0SecondChildTriangle],
					-1,
					-1);

				if (hasParent1)
				{
					parentsFirstChildTriangles.zw = ivec2(
						firstChildTriangleBuffer[parent1FirstChildTriangle],
						firstChildTriangleBuffer[parent1SecondChildTriangle]);
				}

				// Only merge triangles without children
				if (any(greaterThan(parentsFirstChildTriangles, ivec4(-1))))
				{
					return;
				}

				int parent0Sibling0SubdivisionFlags = int(subdivisionFlagsBuffer[parent0FirstChildTriangle]);
				int parent0Sibling1SubdivisionFlags = int(subdivisionFlagsBuffer[parent0SecondChildTriangle]);
				int parent1Sibling0SubdivisionFlags = 0;
				int parent1Sibling1SubdivisionFlags = 0;

				if (hasParent1)
				{
					parent1Sibling0SubdivisionFlags = int(subdivisionFlagsBuffer[parent1FirstChildTriangle]);
					parent1Sibling1SubdivisionFlags = int(subdivisionFlagsBuffer[parent1SecondChildTriangle]);
				}

				int mergeFlags = parent0Sibling0SubdivisionFlags & parent0Sibling1SubdivisionFlags &
					((hasParent1) ? (parent1Sibling0SubdivisionFlags & parent1Sibling1SubdivisionFlags) : FLAG_MERGE);

				// Only merge if all four involved triangles can be merged
				if ((mergeFlags & FLAG_MERGE) == 0)
				{
					return;
				}

				int parent0SubdivisionFlags = int(subdivisionFlagsBuffer[parent0TriangleIndex]);
				int parent1SubdivisionFlags = 0;
				int parent0Neighbor0SubdivisionFlags = 0;
				int parent0Neighbor2SubdivisionFlags = 0;
				int parent1Neighbor0SubdivisionFlags = 0;
				int parent1Neighbor2SubdivisionFlags = 0;

				if (hasParent1)
				{
					parent1SubdivisionFlags = int(subdivisionFlagsBuffer[parent1TriangleIndex]);
				}

				if (parent0AdjacentTriangles[0] > -1)
				{
					parent0Neighbor0SubdivisionFlags = int(subdivisionFlagsBuffer[parent0AdjacentTriangles[0]]);
				}

				if (parent0AdjacentTriangles[2] > -1)
				{
					parent0Neighbor2SubdivisionFlags = int(subdivisionFlagsBuffer[parent0AdjacentTriangles[2]]);
				}

				if (parent1AdjacentTriangles[0] > -1)
				{
					parent1Neighbor0SubdivisionFlags = int(subdivisionFlagsBuffer[parent1AdjacentTriangles[0]]);
				}

				if (parent1AdjacentTriangles[2] > -1)
				{
					parent1Neighbor2SubdivisionFlags = int(subdivisionFlagsBuffer[parent1AdjacentTriangles[2]]);
				}

				// Only merge triangles if their parents do not need to be subdivided, otherwise
				// we end up in a merge-subdivision loop
				if (((parent0SubdivisionFlags | parent1SubdivisionFlags | parent0Neighbor0SubdivisionFlags |
					parent0Neighbor2SubdivisionFlags | parent1Neighbor0SubdivisionFlags | parent1Neighbor2SubdivisionFlags) & FLAG_SUBDIVIDE) != 0)
				{
					return;
				}

				// In general, we only want to process triangles of the current graph color, which one
				// exception of triangles at the borders of a non-watertight mesh, i.e., triangles that
				// do not have a neighbor sharing the hypothenuse. If the non-existent neighbor would
				// have had the current graph color, we need to process these current triangles also in
				// this pass.
				int parent0TriangleGraphColor = int(triangleGraphColorBuffer[parent0TriangleIndex]);
				int graphColor = (hasParent1) ? parent0TriangleGraphColor : (parent0TriangleGraphColor & ~1);

				if ((graphColor & 1) != 0)
				{
					return;
				}

				if (graphColor == 0)
				{
					graphColor0Buffer[atomicAdd(graphColor0Counter, 1)] = currentTriangleIndex;
				}
				else if (graphColor == 2)
				{
					graphColor1Buffer[atomicAdd(graphColor1Counter, 1)] = currentTriangleIndex;
				}
				else if (graphColor == 4)
				{
					graphColor2Buffer[atomicAdd(graphColor2Counter, 1)] = currentTriangleIndex;
				}
				else
				{
					graphColor3Buffer[atomicAdd(graphColor3Counter, 1)] = currentTriangleIndex;
				}
			}
		)
	};

	const std::vector<const char*> MergeCompute
	{
		MULTILINE(
			#version 450\n
			#extension GL_NV_gpu_shader5 : enable\n

			layout(local_size_x = 32, local_size_y = 1, local_size_z = 1) in;

			layout(std430) restrict writeonly buffer subdivisionFlagsBufferBlock
			{
				uint8_t subdivisionFlagsBuffer[];
			};

			layout(std430) restrict buffer firstChildTriangleBufferBlock
			{
				int firstChildTriangleBuffer[];
			};

			layout(std430) restrict readonly buffer parentTriangleBufferBlock
			{
				int parentTriangleBuffer[];
			};

			layout(std430) restrict buffer adjacentTrianglesBufferBlock
			{
				int adjacentTrianglesBuffer[];
			};

			ivec3 adjacentTrianglesBuffer_fetch(int elementIndex)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				return ivec3(adjacentTrianglesBuffer[elementOffsets.x], adjacentTrianglesBuffer[elementOffsets.y], adjacentTrianglesBuffer[elementOffsets.z]);
			}

			void adjacentTrianglesBuffer_storeComponent(int elementIndex, int componentIndex, int value)
			{
				adjacentTrianglesBuffer[3 * elementIndex + componentIndex] = value;
			}

			layout(std430) restrict readonly buffer vertexIndicesBufferBlock
			{
				int vertexIndicesBuffer[];
			};

			int vertexIndicesBuffer_fetchX(int elementIndex)
			{
				return vertexIndicesBuffer[3 * elementIndex];
			}

			layout(std430) restrict writeonly buffer freeTriangleIndicesBufferBlock
			{
				int freeTriangleIndicesBuffer[];
			};

			layout(std430) restrict writeonly buffer freeVertexIndicesBufferBlock
			{
				int freeVertexIndicesBuffer[];
			};

			layout(std430) restrict readonly buffer graphColorBufferBlock
			{
				int graphColorBuffer[];
			};

			uniform int freeTriangleIndicesOffset;
			uniform int freeVertexIndicesOffset;
			uniform int numThreads;
		),
		CommonCornelEtAl2022,
		MULTILINE(
			int getIndexOfValue(ivec3 vec, int value)
			{
				ivec3 hasValue = ivec3(equal(vec, ivec3(value))) << ivec3(0, 1, 2);

				return findLSB(hasValue.x + hasValue.y + hasValue.z);
			}

			void main()
			{
				int threadIndex = int(gl_GlobalInvocationID.x);

				if (threadIndex >= numThreads)
				{
					return;
				}

				int currentTriangleIndex = graphColorBuffer[threadIndex];

				int parent0TriangleIndex = parentTriangleBuffer[currentTriangleIndex];
				ivec3 parent0AdjacentTriangles = adjacentTrianglesBuffer_fetch(parent0TriangleIndex);
				int parent1TriangleIndex = parent0AdjacentTriangles[1];
				bool hasParent1 = (parent1TriangleIndex > -1);

				int parent0FirstChildTriangle = firstChildTriangleBuffer[parent0TriangleIndex];
				int parent0SecondChildTriangle = parent0FirstChildTriangle + 1;
				int parent1FirstChildTriangle = -1;
				int parent1SecondChildTriangle = -1;

				if (hasParent1)
				{
					parent1FirstChildTriangle = firstChildTriangleBuffer[parent1TriangleIndex];
					parent1SecondChildTriangle = parent1FirstChildTriangle + 1;
				}

				// Merging consists of four steps
				// 1: Flag the child triangles as deleted
				// 2: Remove the child indices from the parent triangles
				// 3: Update the adjacency indices of the parent triangles
				// 4: Update the adjacency indices of the adjacent triangles

				freeTriangleIndicesBuffer[freeTriangleIndicesOffset + threadIndex] = parent0FirstChildTriangle;
				freeVertexIndicesBuffer[freeVertexIndicesOffset + threadIndex] = vertexIndicesBuffer_fetchX(parent0FirstChildTriangle);

				ivec3 parent0Sibling0AdjacentTriangles = adjacentTrianglesBuffer_fetch(parent0FirstChildTriangle);
				ivec3 parent0Sibling1AdjacentTriangles = adjacentTrianglesBuffer_fetch(parent0SecondChildTriangle);
				ivec3 parent1Sibling0AdjacentTriangles = ivec3(-1);
				ivec3 parent1Sibling1AdjacentTriangles = ivec3(-1);

				if (hasParent1)
				{
					parent1Sibling0AdjacentTriangles = adjacentTrianglesBuffer_fetch(parent1FirstChildTriangle);
					parent1Sibling1AdjacentTriangles = adjacentTrianglesBuffer_fetch(parent1SecondChildTriangle);
				}

				firstChildTriangleBuffer[parent0TriangleIndex] = -1;
				adjacentTrianglesBuffer_storeComponent(parent0TriangleIndex, 2, parent0Sibling0AdjacentTriangles[1]);
				adjacentTrianglesBuffer_storeComponent(parent0TriangleIndex, 0, parent0Sibling1AdjacentTriangles[1]);

				if (hasParent1)
				{
					firstChildTriangleBuffer[parent1TriangleIndex] = -1;
					adjacentTrianglesBuffer_storeComponent(parent1TriangleIndex, 2, parent1Sibling0AdjacentTriangles[1]);
					adjacentTrianglesBuffer_storeComponent(parent1TriangleIndex, 0, parent1Sibling1AdjacentTriangles[1]);
				}

				ivec3 parent0Sibling0NeighborAdjacentTriangles = ivec3(-1);
				ivec3 parent0Sibling1NeighborAdjacentTriangles = ivec3(-1);
				ivec3 parent1Sibling0NeighborAdjacentTriangles = ivec3(-1);
				ivec3 parent1Sibling1NeighborAdjacentTriangles = ivec3(-1);

				if (parent0Sibling0AdjacentTriangles[1] > -1)
				{
					parent0Sibling0NeighborAdjacentTriangles = adjacentTrianglesBuffer_fetch(parent0Sibling0AdjacentTriangles[1]);
				}

				if (parent0Sibling1AdjacentTriangles[1] > -1)
				{
					parent0Sibling1NeighborAdjacentTriangles = adjacentTrianglesBuffer_fetch(parent0Sibling1AdjacentTriangles[1]);
				}

				if (parent1Sibling0AdjacentTriangles[1] > -1)
				{
					parent1Sibling0NeighborAdjacentTriangles = adjacentTrianglesBuffer_fetch(parent1Sibling0AdjacentTriangles[1]);
				}

				if (parent1Sibling1AdjacentTriangles[1] > -1)
				{
					parent1Sibling1NeighborAdjacentTriangles = adjacentTrianglesBuffer_fetch(parent1Sibling1AdjacentTriangles[1]);
				}

				int parent0FirstChildTriangleIndex = getIndexOfValue(parent0Sibling0NeighborAdjacentTriangles, parent0FirstChildTriangle);
				int parent0SecondChildTriangleIndex = getIndexOfValue(parent0Sibling1NeighborAdjacentTriangles, parent0SecondChildTriangle);

				if (parent0FirstChildTriangleIndex > -1)
				{
					adjacentTrianglesBuffer_storeComponent(parent0Sibling0AdjacentTriangles[1], parent0FirstChildTriangleIndex, parent0TriangleIndex);
				}

				if (parent0SecondChildTriangleIndex > -1)
				{
					adjacentTrianglesBuffer_storeComponent(parent0Sibling1AdjacentTriangles[1], parent0SecondChildTriangleIndex, parent0TriangleIndex);
				}

				if (hasParent1)
				{
					int parent1FirstChildTriangleIndex = getIndexOfValue(parent1Sibling0NeighborAdjacentTriangles, parent1FirstChildTriangle);
					int parent1SecondChildTriangleIndex = getIndexOfValue(parent1Sibling1NeighborAdjacentTriangles, parent1SecondChildTriangle);

					if (parent1FirstChildTriangleIndex > -1)
					{
						adjacentTrianglesBuffer_storeComponent(parent1Sibling0AdjacentTriangles[1], parent1FirstChildTriangleIndex, parent1TriangleIndex);
					}

					if (parent1SecondChildTriangleIndex > -1)
					{
						adjacentTrianglesBuffer_storeComponent(parent1Sibling1AdjacentTriangles[1], parent1SecondChildTriangleIndex, parent1TriangleIndex);
					}
				}

				subdivisionFlagsBuffer[parent0FirstChildTriangle] = uint8_t(FLAG_DELETED);
				subdivisionFlagsBuffer[parent0SecondChildTriangle] = uint8_t(FLAG_DELETED);

				if (hasParent1)
				{
					subdivisionFlagsBuffer[parent1FirstChildTriangle] = uint8_t(FLAG_DELETED);
					subdivisionFlagsBuffer[parent1SecondChildTriangle] = uint8_t(FLAG_DELETED);
				}
			}
		)
	};

	const std::vector<const char*> RenderingVertex
	{
		MULTILINE(
			#version 450\n

			in vec4 inPosition;
		),
		CameraUniforms,
		MULTILINE(
			void main()
			{
				gl_Position = viewProjectionMatrix * inPosition;
			}
		)
	};
}

#endif