#pragma once
#ifndef TessellationCornelEtAl2022_h
#define TessellationCornelEtAl2022_h

#include "../AbstractTessellation.h"
#include "../Program.h"

// This struct is defined by OpenGL, do not change it!
struct DrawElementsIndirectCommand
{
	DrawElementsIndirectCommand() :
		primitiveCount(0u),
		instanceCount(0u),
		firstIndex(0u),
		baseVertex(0u),
		baseInstance(0u)
	{ }

	DrawElementsIndirectCommand(
		GLuint primitiveCount,
		GLuint instanceCount,
		GLuint firstIndex,
		GLuint baseVertex,
		GLuint baseInstance) :
		primitiveCount(primitiveCount),
		instanceCount(instanceCount),
		firstIndex(firstIndex),
		baseVertex(baseVertex),
		baseInstance(baseInstance)
	{ }

	GLuint primitiveCount;
	GLuint instanceCount;
	GLuint firstIndex;
	GLuint baseVertex;
	GLuint baseInstance;
};

//=================================================================================================
/**
*	This class implements watertight incremental heightfield tessellation.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
class TessellationCornelEtAl2022 : public AbstractTessellation
{
public:
	/// Constructor
	TessellationCornelEtAl2022(const Heightfield& heightfield);
	/// Destructor
	~TessellationCornelEtAl2022() override;
	/// Performs a per-frame update of the tessellation
	void update(const Camera& camera) override;
	/// Renders the geometry
	void render(const Camera& camera) override;

private:
	/// Updates the subdivision flags of all triangles based on their distance to the camera
	void updateSubdivisionFlags(const Camera& camera);
	/// Subdivides all triangles flagged for subdivision
	bool subdivideTriangles(const Camera& camera, int& outNumSelectedTriangles, int& outRequiredResizeElements);
	/// Merges all triangles flagged for merge
	bool mergeTriangles(const Camera& camera, int& outNumSelectedTriangles);
	/// Compact all currently visible triangles for draw call
	void compactTriangles(const Camera& camera);

public:
	GLuint mSubdivisionFlagsBuffer = 0u;
	GLuint mParentTriangleBuffer = 0u;
	GLuint mFirstChildTriangleBuffer = 0u;
	GLuint mAdjacentTrianglesBuffer = 0u;
	GLuint mVertexIndicesBuffer = 0u;
	GLuint mTriangleGraphColorBuffer = 0u;
	GLuint mVertexBuffer = 0u;
	GLuint mFreeTriangleIndicesBuffer = 0u;
	GLuint mFreeVertexIndicesBuffer = 0u;

	std::vector<GLuint> mGraphColorBuffers { 0u, 0u, 0u, 0u };
	std::vector<GLuint> mGraphColorCounters { 0u, 0u, 0u, 0u };
	GLuint mIndirectDrawBuffer = 0u;
	GLuint mVisibleTriangleBuffer = 0u;
	GLuint mVAO = 0u;

	Eigen::Matrix4f mPreviousViewProjectionMatrix = Eigen::Matrix4f::Zero();

	std::unique_ptr<Program> mUpdateSubdivisionFlagsProgram = nullptr;
	std::unique_ptr<Program> mSubdivisionTriangleSelectionProgram = nullptr;
	std::unique_ptr<Program> mSubdivisionProgram = nullptr;
	std::unique_ptr<Program> mMergeTriangleSelectionProgram = nullptr;
	std::unique_ptr<Program> mMergeProgram = nullptr;
	std::unique_ptr<Program> mCompactionProgram = nullptr;
	std::unique_ptr<Program> mRenderingProgram = nullptr;

	int mNumReusableTriangleIndices = 0;
	int mNumReusableVertexIndices = 0;
	int mUnusedTrianglesOffset = 0;
	int mUnusedVerticesOffset = 0;

	bool mFirstFrame = true;	///< Flag to force a complete tessellation no matter the time budget
};

#endif