#include "Heightfield.h"

#include "CommonShaders.h"

using namespace Eigen;

//=================================================================================================
/**
*	Destructor.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
Heightfield::Heightfield(const Vector2i& dimensions, const Vector2f& cellSize, const Vector2f& origin, float scaleFactor) :
	mDimensions(dimensions),
	mCellSize(cellSize),
	mOrigin(origin),
	mScaleFactor(scaleFactor)
{
	mAABB.extend(Vector3f(mOrigin.x(), mOrigin.y(), 0.0f));
	mAABB.extend(Vector3f(mOrigin.x() + mDimensions.x() * mCellSize.x(), mOrigin.y() + mDimensions.y() * mCellSize.y(), mScaleFactor));

	//=================================================================================================
	/**
	*	Create chunk descriptor for tessellation
	*/
	//=================================================================================================

	double divisor = 1.0;
	double estimatedNumCells = mDimensions.x() * mDimensions.y();

	while (estimatedNumCells > static_cast<double>(mChunkDescriptor.mDesiredNumChunks))
	{
		divisor *= 2.0;

		estimatedNumCells = std::floor(mDimensions.x() / std::round(divisor)) * std::floor(mDimensions.y() / std::round(divisor));
	}

	if (estimatedNumCells == static_cast<double>(mChunkDescriptor.mDesiredNumChunks))
	{
		mChunkDescriptor.mChunkDivisor = static_cast<int>(std::round(divisor));
	}
	else
	{
		mChunkDescriptor.mChunkDivisor = std::max(1, static_cast<int>(std::round(divisor / 2.0)));
	}

	mChunkDescriptor.mChunkGridDimensions = Vector2i(
		mDimensions.x() / mChunkDescriptor.mChunkDivisor,
		mDimensions.y() / mChunkDescriptor.mChunkDivisor);

	if (mChunkDescriptor.mChunkGridDimensions.x() * mChunkDescriptor.mChunkDivisor < mDimensions.x())
	{
		mChunkDescriptor.mChunkGridDimensions.x()++;
	}

	if (mChunkDescriptor.mChunkGridDimensions.y() * mChunkDescriptor.mChunkDivisor < mDimensions.y())
	{
		mChunkDescriptor.mChunkGridDimensions.y()++;
	}

	mChunkDescriptor.mNumChunks = mChunkDescriptor.mChunkGridDimensions.x() * mChunkDescriptor.mChunkGridDimensions.y();
	mChunkDescriptor.mChunkCellSize = mCellSize * static_cast<float>(mChunkDescriptor.mChunkDivisor);

	//=================================================================================================
	/**
	*	Create a heightfield buffer from a noise function
	*/
	//=================================================================================================

	glCreateBuffers(1, &mBuffer);
	glNamedBufferData(mBuffer, sizeof(GLfloat) * dimensions.x() * dimensions.y(), nullptr, GL_STATIC_DRAW);

	mNoiseProgram = std::make_unique<Program>(NoiseCompute);
	mNoiseProgram->bind();
	mNoiseProgram->setUniform1f("heightfieldScaleFactor", mScaleFactor);
	mNoiseProgram->setUniform2f("cellSize", mCellSize);
	mNoiseProgram->setUniform2f("domainOrigin", mOrigin);
	mNoiseProgram->setUniform2i("gridDimensions", mDimensions);
	mNoiseProgram->setSSBO("heightfieldBuffer", mBuffer);
	mNoiseProgram->compute1D(dimensions.x() * dimensions.y());
	mNoiseProgram->unbind();

	glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
}

//=================================================================================================
/**
*	Destructor.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
Heightfield::~Heightfield()
{
	glDeleteBuffers(1, &mBuffer);
}