#pragma once
#ifndef ShaderCodesKerblEtAl2022_h
#define ShaderCodesKerblEtAl2022_h

namespace ShaderCodesKerblEtAl2022
{
#include "../CommonShaders.h"

	const char* CommonDefines = MULTILINE(
		#version 450\n
		#extension GL_EXT_gpu_shader4 : require\n

		const uint bufferCapacity = 1u << 24u; // Note: This value is also defined in the cpp

		// This struct is defined by OpenGL
		struct DispatchIndirectCommand
		{
			uint num_groups_x;
			uint num_groups_y;
			uint num_groups_z;
		};

		// This struct is defined by OpenGL
		struct DrawElementsIndirectCommand
		{
			uint count;
			uint primCount;
			uint firstIndex;
			uint baseVertex;
			uint baseInstance;
		};
	);

	//=================================================================================================
	/**
	*	This code is based on isubd.glsl - public domain implicit subdivision on the GPU (created by
	*	Jonathan Dupuy) and the extensions provided by Linus Horvath and Bernhard Kerbl. See An
	*	Improved Triangle Encoding Scheme for Cached Tessellation. Bernhard Kerbl, Linus Horvath,
	*	Daniel Cornel, and Michael Wimmer. Eurographics 2022 - Short Papers, pages 53-56, 2022.
	*/
	//=================================================================================================
	const char* ISUBD = MULTILINE(
		uint encodeKey(in uint z, in uint zBits, in uint x, in uint y, in uint xyBits)
		{
			//                  level                             z                     x         y
			return (1u << (zBits + xyBits + xyBits)) | (z << (xyBits + xyBits)) | (x << xyBits) | y;
		}

		void decodeKey(in uint key, out uint z, out uint zBits, out uint x, out uint y, out uint xyBits)
		{
			uint level = findMSB(key);

			zBits = ((level + 1u) & 1u) + 1u;
			xyBits = (level - 1u) >> 1u;

			uint yMask = (1u << xyBits) - 1u;
			y = key & yMask;

			// xMask = (yMask << xyBits)
			x = (key & (yMask << xyBits)) >> xyBits;

			// zMask = (((1u << zBits) - 1u) << (xyBits + xyBits))
			z = (key & (((1u << zBits) - 1u) << (xyBits + xyBits))) >> (xyBits + xyBits);
		}

		uint parentKey(in uint key)
		{
			uint z, zBits, x, y, xyBits; decodeKey(key, z, zBits, x, y, xyBits);

			// Readable version
			/*bool isFlipped = ((x + y) & 1u) == 0u;
			if (zBits == 1u) {
				if (isFlipped) {
					z = z | ((x & 1u) << 1u);
				} else {
					z = (z << 1u) | (x & 1u);
				}

				zBits = 2u;
				x = x >> 1u;
				y = y >> 1u;
				xyBits = xyBits - 1u;
			} else {
				if (isFlipped) {
					z = z & 1u;
				} else {
					z = z >> 1u;
				}

				zBits = 1u;
			}*/

			// Arithmetic version
			uint zBitsIsOne = zBits & 1u; // 1 if |\| or |/|, otherwise 0
			uint zBitsIsTwo = zBitsIsOne ^ 1u; // 1 if |X|, otherwise 0

			uint isOrientedNormally = (x + y) & 1u; // 1 if |\|, otherwise 0
			uint isFlipped = isOrientedNormally ^ 1u; // 1 if |/|, otherwise 0

			z = (zBitsIsOne * ((z << isOrientedNormally) | ((x & 1u) << isFlipped))) +
				(zBitsIsTwo * (
				(isFlipped * (z & 1u)) +
					(isOrientedNormally * (z >> 1u))));
			zBits = 1u + (zBitsIsOne * 1u);

			x = (zBitsIsOne * (x >> 1u)) +
				(zBitsIsTwo * x);
			y = (zBitsIsOne * (y >> 1u)) +
				(zBitsIsTwo * y);
			xyBits = xyBits - (zBitsIsOne * 1u);

			return encodeKey(z, zBits, x, y, xyBits);
		}

		void childrenKeys(in uint key, out uint children[2])
		{
			uint z, zBits, x, y, xyBits; decodeKey(key, z, zBits, x, y, xyBits);

			// Readable version
			/*uint x1, y1, z1;
			uint x2, y2, z2;

			bool isFlipped = ((x + y) & 1u) == 0u;
			if (zBits == 1u) {
				x1 = x;
				y1 = y;
				x2 = x;
				y2 = y;

				if (isFlipped) {
					z1 = (0u << 1u) | z;
					z2 = (1u << 1u) | z;
				} else {
					z1 = z << 1u;
					z2 = (z << 1u) | 1u;
				}

				zBits = 2u;
			} else { // zBits == 2u
				if (z == 0u) {
					x1 = x * 2u;
					y1 = (y * 2u) + 1u;

					x2 = x * 2u;
					y2 = y * 2u;
				} else if (z == 1u) {
					x1 = x * 2u;
					y1 = y * 2u;

					x2 = (x * 2u) + 1u;
					y2 = y * 2u;
				} else if (z == 2u) {
					x1 = x * 2u;
					y1 = (y * 2u) + 1u;

					x2 = (x * 2u) + 1u;
					y2 = (y * 2u) + 1u;
				} else { // z == 3u
					x1 = (x * 2u) + 1u;
					y1 = (y * 2u) + 1u;

					x2 = (x * 2u) + 1u;
					y2 = y * 2u;
				}
				xyBits = xyBits + 1u;

				bool isSameX = (x2 - x1) == 0u;
				bool isSameY = (y1 - y2) == 0u;
				if (isSameX && !isSameY) {
					z1 = x1 & 1u;
					z2 = x2 & 1u;
				} else if (!isSameX && isSameY) {
					z1 = (x1 + 1u) & 1u;
					z2 = (x2 + 1u) & 1u;
				}

				zBits = 1u;
			}*/

			// Arithmetic version
			uint zBitsIsOne = zBits & 1u; // 1 if |\| or |/|, otherwise 0
			uint zBitsIsTwo = zBitsIsOne ^ 1u; // 1 if |X|, otherwise 0

			uint isOrientedNormally = (x + y) & 1u; // 1 if |\|, otherwise 0
			uint isFlipped = isOrientedNormally ^ 1u; // 1 if |/|, otherwise 0

			uint zIsZero = uint(z == 0u);
			uint zIsOne = uint(z == 1u);
			uint zIsTwo = uint(z == 2u);
			uint zIsThree = uint(z == 3u);
			uint xNewBase = x * 2u;
			uint yNewBase = y * 2u;

			uint x1 = (zBitsIsOne * x) +
				(zBitsIsTwo * (
				((zIsZero + zIsOne + zIsTwo) * xNewBase) +
					(zIsThree * (xNewBase + 1u))));
			uint y1 = (zBitsIsOne * y) +
				(zBitsIsTwo * (
				((zIsZero + zIsTwo + zIsThree) * (yNewBase + 1u)) +
					(zIsOne * yNewBase)));
			uint x2 = (zBitsIsOne * x) +
				(zBitsIsTwo * (
				(zIsZero * xNewBase) +
					((zIsOne + zIsTwo + zIsThree) * (xNewBase + 1u))));
			uint y2 = (zBitsIsOne * y) +
				(zBitsIsTwo * (
				((zIsZero + zIsOne + zIsThree) * yNewBase) +
					(zIsTwo * (yNewBase + 1u))));
			xyBits = xyBits + (zBitsIsTwo * 1u);

			uint isNotSameX = x2 - x1; // 1 if x1 != x2, otherwise 0
			uint isSameX = isNotSameX ^ 1u; // 1 if x1 == x2, otherwise 0
			uint isNotSameY = y1 - y2; // 1 if y1 != y2, otherwise 0
			uint isSameY = isNotSameY ^ 1u; // 1 if y1 == y2, otherwise 0

			uint z1 = (zBitsIsOne * (
				(isOrientedNormally * (z << 1u)) +
				(isFlipped * ((0u << 1u) | z)))) +
				(zBitsIsTwo * (
				(isSameX * isNotSameY * (x1 & 1u)) +
					(isNotSameX * isSameY * ((x1 + 1u) & 1u))));
			uint z2 = (zBitsIsOne * (
				(isOrientedNormally * ((z << 1u) | 1u)) +
				(isFlipped * ((1u << 1u) | z)))) +
				(zBitsIsTwo * (
				(isSameX * isNotSameY * (x2 & 1u)) +
					(isNotSameX * isSameY * ((x2 + 1u) & 1u))));
			zBits = 1u + (zBitsIsOne * 1u);

			children[0] = encodeKey(z1, zBits, x1, y1, xyBits);
			children[1] = encodeKey(z2, zBits, x2, y2, xyBits);
		}

		bool isRootKey(in uint key)
		{
			return (key <= 3u); // binary 10 or 11, 0 or 1 being the required id
		}

		bool isLeafKey(in uint key)
		{
			return findMSB(key) == 31;
		}

		bool isChildZeroKey(in uint key)
		{
			uint z, zBits, x, y, xyBits; decodeKey(key, z, zBits, x, y, xyBits);

			// Readable version
			bool isLeft = (x & 1u) == 0u;
			bool isRight = !isLeft;
			bool isUp = (y & 1u) == 1u;
			bool isDown = !isUp;

			if (zBits == 1u)
			{
				if (isLeft && isUp)
				{
					return true;
				}
				else if (isLeft && isDown && z == 1u)
				{
					return true;
				}
				else if (isRight && isUp && z == 1u)
				{
					return true;
				}
			}
			else if (zBits == 2u)
			{
				if (z == 0u || z == 3u)
				{
					return true;
				}
			}

			// Arithmetic version
			// How?

			return false;
		}

		void calculateBarycentricCoordinates(in uint key, out vec2 u_out[3])
		{
			uint z, zBits, x, y, xyBits; decodeKey(key, z, zBits, x, y, xyBits);

			float stepSize = 1.0 / pow(2u, xyBits);
			float lowerBoundX = stepSize * x;
			float lowerBoundY = stepSize * y;
			float upperBoundX = lowerBoundX + stepSize;
			float upperBoundY = lowerBoundY + stepSize;

			vec2 pointLowerLeft = vec2(lowerBoundX, lowerBoundY);
			vec2 pointLowerRight = vec2(upperBoundX, lowerBoundY);
			vec2 pointUpperRight = vec2(upperBoundX, upperBoundY);
			vec2 pointUpperLeft = vec2(lowerBoundX, upperBoundY);
			vec2 pointCenter = vec2(lowerBoundX + (stepSize * 0.5f), lowerBoundY + (stepSize * 0.5f));

			// Readable version
			/*if (zBits == 1u) {
				bool isFlipped = ((x + y) & 1u) == 0u;
				if (isFlipped) { // if |/|
					if (z == 0u) {
						u_out[0] = pointUpperLeft;
						u_out[1] = pointUpperRight;
						u_out[2] = pointLowerLeft;
					} else { // z == 1u
						u_out[0] = pointLowerRight;
						u_out[1] = pointLowerLeft;
						u_out[2] = pointUpperRight;
					}
				} else { // if |\|
					if (z == 0u) {
						u_out[0] = pointLowerLeft;
						u_out[1] = pointUpperLeft;
						u_out[2] = pointLowerRight;
					} else { // z == 1u
						u_out[0] = pointUpperRight;
						u_out[1] = pointLowerRight;
						u_out[2] = pointUpperLeft;
					}
				}
			} else { // zBits == 2u
				if (z == 0u) {
					u_out[0] = pointCenter;
					u_out[1] = pointLowerLeft;
					u_out[2] = pointUpperLeft;
				} else if (z == 1u) {
					u_out[0] = pointCenter;
					u_out[1] = pointLowerRight;
					u_out[2] = pointLowerLeft;
				} else if (z == 2u) {
					u_out[0] = pointCenter;
					u_out[1] = pointUpperLeft;
					u_out[2] = pointUpperRight;
				} else { // z == 3u
					u_out[0] = pointCenter;
					u_out[1] = pointUpperRight;
					u_out[2] = pointLowerRight;
				}
			}*/

			// Arithmetic version
			uint zBitsIsOne = zBits & 1u; // 1 if |\| or |/|, otherwise 0
			uint zBitsIsTwo = zBitsIsOne ^ 1u; // 1 if |X|, otherwise 0

			uint isOrientedNormally = (x + y) & 1u; // 1 if |\|, otherwise 0
			uint isFlipped = isOrientedNormally ^ 1u; // 1 if |/|, otherwise 0

			uint zIsZero = uint(z == 0u);
			uint zIsOne = uint(z == 1u);
			uint zIsTwo = uint(z == 2u);
			uint zIsThree = uint(z == 3u);

			u_out[0] = (zBitsIsOne * (
				(isFlipped * (
				(zIsZero * pointUpperLeft) +
					(zIsOne * pointLowerRight))) +
					(isOrientedNormally * (
				(zIsZero * pointLowerLeft) +
						(zIsOne * pointUpperRight))))) +
						(zBitsIsTwo * pointCenter);
			u_out[1] = (zBitsIsOne * (
				(isFlipped * (
				(zIsZero * pointUpperRight) +
					(zIsOne * pointLowerLeft))) +
					(isOrientedNormally * (
				(zIsZero * pointUpperLeft) +
						(zIsOne * pointLowerRight))))) +
						(zBitsIsTwo * (
				(zIsZero * pointLowerLeft) +
							(zIsOne * pointLowerRight) +
							(zIsTwo * pointUpperLeft) +
							(zIsThree * pointUpperRight)));
			u_out[2] = (zBitsIsOne * (
				(isFlipped * (
				(zIsZero * pointLowerLeft) +
					(zIsOne * pointUpperRight))) +
					(isOrientedNormally * (
				(zIsZero * pointLowerRight) +
						(zIsOne * pointUpperLeft))))) +
						(zBitsIsTwo * (
				(zIsZero * pointUpperLeft) +
							(zIsOne * pointLowerLeft) +
							(zIsTwo * pointUpperRight) +
							(zIsThree * pointLowerRight)));
		}

		// barycentric interpolation
		vec3 berp(in vec3 v[3], in vec2 u)
		{
			return v[0] + u.x * (v[1] - v[0]) + u.y * (v[2] - v[0]);
		}
		vec4 berp(in vec4 v[3], in vec2 u)
		{
			return v[0] + u.x * (v[1] - v[0]) + u.y * (v[2] - v[0]);
		}

		// subdivision routine (vertex position only)
		void subd(in uint key, in vec4 v_in[3], out vec4 v_out[3])
		{
			vec2 u[3]; calculateBarycentricCoordinates(key, u);

			v_out[0] = berp(v_in, u[0]);
			v_out[1] = berp(v_in, u[1]);
			v_out[2] = berp(v_in, u[2]);
		}

		void subd(in uint key, in vec3 v_in[3], out vec3 v_out[3])
		{
			vec2 u[3]; calculateBarycentricCoordinates(key, u);

			v_out[0] = berp(v_in, u[0]);
			v_out[1] = berp(v_in, u[1]);
			v_out[2] = berp(v_in, u[2]);
		}

		// subdivision routine (vertex position only)
		// also computes parent position
		void subd(in uint key, in vec4 v_in[3], out vec4 v_out[3], out vec4 v_out_p[3])
		{
			vec2 u[3]; calculateBarycentricCoordinates(key, u);
			vec2 uParent[3]; calculateBarycentricCoordinates(parentKey(key), uParent);

			v_out[0] = berp(v_in, u[0]);
			v_out[1] = berp(v_in, u[1]);
			v_out[2] = berp(v_in, u[2]);

			v_out_p[0] = berp(v_in, uParent[0]);
			v_out_p[1] = berp(v_in, uParent[1]);
			v_out_p[2] = berp(v_in, uParent[2]);
		}
		void subd(in uint key, in vec3 v_in[3], out vec3 v_out[3], out vec3 v_out_p[3])
		{
			vec2 u[3]; calculateBarycentricCoordinates(key, u);
			vec2 uParent[3]; calculateBarycentricCoordinates(parentKey(key), uParent);

			v_out[0] = berp(v_in, u[0]);
			v_out[1] = berp(v_in, u[1]);
			v_out[2] = berp(v_in, u[2]);

			v_out_p[0] = berp(v_in, uParent[0]);
			v_out_p[1] = berp(v_in, uParent[1]);
			v_out_p[2] = berp(v_in, uParent[2]);
		}

		layout(std430) restrict writeonly buffer outputSubdivisionBufferBlock
		{
			uvec2 outputSubdivisionBuffer[];
		};

		layout(std430) restrict buffer numOutputTrianglesBufferBlock
		{
			uint numOutputTrianglesBuffer;
		};

		void writeKey(uint primID, uint key)
		{
			uint idx = atomicAdd(numOutputTrianglesBuffer, 1u);

			if (idx < bufferCapacity)
			{
				outputSubdivisionBuffer[idx] = uvec2(primID, key);
			}
			else
			{
				// Note: We discard all triangles that do not fit in the output buffer.
				// This will inevitably break the triangulation. A proper mechanism is
				// needed to guarantee that parent triangles will be stored in case a
				// child triangle cannot be written anymore.

				const uint lastIndex = bufferCapacity - 1u;

				atomicExchange(numOutputTrianglesBuffer, lastIndex);
			}
		}

		bool updateSubdBuffer(uint primID, uint key, int targetLod, int parentLod)
		{
			int keyLod = findMSB(key) - 1;

			if (keyLod < targetLod && !isLeafKey(key))
			{
				uint children[2];
				childrenKeys(key, children);

				writeKey(primID, children[0]);
				writeKey(primID, children[1]);

				return true;
			}
			else if (keyLod < (parentLod + 1))
			{
				writeKey(primID, key);
			}
			else
			{
				if (isRootKey(key))
				{
					writeKey(primID, key);
				}
				else if (isChildZeroKey(key))
				{
					writeKey(primID, parentKey(key));

					return true;
				}
			}

			return false;
		}

		uniform float lodFactor;

		float distanceToLod(float z, float lodFactor)
		{
			float lod = -2.0 * log2(clamp(z * lodFactor, 0.0, 1.0));

			return min(lod, 30.0);
		}

		float computeLod(vec3 c)
		{
			c.z += sampleHeightfield(inverseViewMatrix[3].xy);

			vec3 cxf = (viewMatrix * vec4(c, 1.0)).xyz;
			float z = length(cxf);

			return distanceToLod(z, lodFactor);
		}

		float computeLod(in vec3 v[3])
		{
			vec3 c = (v[1].xyz + v[2].xyz) * 0.5;

			return computeLod(c);
		}
	);

	const std::vector<const char*> TerrainVertex
	{
		CommonDefines,
		MULTILINE(
			in vec2 inPosition;
		),
		CameraUniforms,
		HeightfieldSampling,
		ISUBD,
		MULTILINE(
			layout(std430) restrict readonly buffer vertexBufferBlock
			{
				float vertexBuffer[];
			};

			vec3 vertexBuffer_fetch(int elementIndex)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				return vec3(vertexBuffer[elementOffsets.x], vertexBuffer[elementOffsets.y], vertexBuffer[elementOffsets.z]);
			}

			layout(std430) restrict readonly buffer indexBufferBlock
			{
				int indexBuffer[];
			};

			ivec3 indexBuffer_fetch(uint elementIndex)
			{
				uvec3 elementOffsets = uvec3(3u * elementIndex) + uvec3(0u, 1u, 2u);

				return ivec3(indexBuffer[elementOffsets.x], indexBuffer[elementOffsets.y], indexBuffer[elementOffsets.z]);
			}

			layout(std430) restrict readonly buffer culledTrianglesBufferBlock
			{
				uvec2 culledTrianglesBuffer[];
			};

			void main()
			{
				int threadID = gl_InstanceID;

				uvec2 triangleEncoding = culledTrianglesBuffer[threadID];

				uint rootTriangleID = triangleEncoding.x;
				ivec3 rootTriangle = indexBuffer_fetch(rootTriangleID);
				vec3 rootTriangleVertices[3] = vec3[3](
					vertexBuffer_fetch(rootTriangle.x),
					vertexBuffer_fetch(rootTriangle.y),
					vertexBuffer_fetch(rootTriangle.z));

				uint triangleSubdivisionKey = triangleEncoding.y;
				vec3 triangleVertices[3];
				subd(triangleSubdivisionKey, rootTriangleVertices, triangleVertices);

				vec3 finalVertex = berp(triangleVertices, inPosition);

				finalVertex.z = sampleHeightfield(finalVertex.xy);

				gl_Position = viewProjectionMatrix * vec4(finalVertex, 1.0);
			}
		)
	};

	const std::vector<const char*> LoDCompute
	{
		CommonDefines,
		CameraUniforms,
		HeightfieldSampling,
		ISUBD,
		MULTILINE(
			layout(local_size_x = 32, local_size_y = 1, local_size_z = 1) in;

			layout(std430) restrict readonly buffer vertexBufferBlock
			{
				float vertexBuffer[];
			};

			vec3 vertexBuffer_fetch(int elementIndex)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				return vec3(vertexBuffer[elementOffsets.x], vertexBuffer[elementOffsets.y], vertexBuffer[elementOffsets.z]);
			}

			layout(std430) restrict readonly buffer indexBufferBlock
			{
				int indexBuffer[];
			};

			ivec3 indexBuffer_fetch(uint elementIndex)
			{
				uvec3 elementOffsets = uvec3(3u * elementIndex) + uvec3(0u, 1u, 2u);

				return ivec3(indexBuffer[elementOffsets.x], indexBuffer[elementOffsets.y], indexBuffer[elementOffsets.z]);
			}

			layout(std430) restrict readonly buffer inputSubdivisionBufferBlock
			{
				uvec2 inputSubdivisionBuffer[];
			};

			layout(std430) restrict writeonly buffer culledTrianglesBufferBlock
			{
				uvec2 culledTrianglesBuffer[];
			};

			layout(std430) restrict readonly buffer numInputTrianglesBufferBlock
			{
				uint numInputTrianglesBuffer;
			};

			layout(std430) restrict buffer indirectDrawBufferBlock
			{
				DrawElementsIndirectCommand indirectDrawBuffer;
			};

			layout(std430) restrict buffer changesAtomicCounterBufferBlock
			{
				int changesAtomicCounterBuffer;
			};

			bool inFrustum(vec3 boxMin, vec3 boxMax)
			{
				vec4 frustumPlanes[6] = vec4[](
					frustumPlaneLeft,
					frustumPlaneRight,
					frustumPlaneBottom,
					frustumPlaneTop,
					frustumPlaneFar,
					frustumPlaneNear);

				for (int i = 0; i < 6; i++)
				{
					int outs = 0;

					outs += ((dot(frustumPlanes[i], vec4(boxMin.x, boxMin.y, boxMin.z, 1.0)) < 0.0) ? 1 : 0);
					outs += ((dot(frustumPlanes[i], vec4(boxMax.x, boxMin.y, boxMin.z, 1.0)) < 0.0) ? 1 : 0);
					outs += ((dot(frustumPlanes[i], vec4(boxMin.x, boxMax.y, boxMin.z, 1.0)) < 0.0) ? 1 : 0);
					outs += ((dot(frustumPlanes[i], vec4(boxMax.x, boxMax.y, boxMin.z, 1.0)) < 0.0) ? 1 : 0);
					outs += ((dot(frustumPlanes[i], vec4(boxMin.x, boxMin.y, boxMax.z, 1.0)) < 0.0) ? 1 : 0);
					outs += ((dot(frustumPlanes[i], vec4(boxMax.x, boxMin.y, boxMax.z, 1.0)) < 0.0) ? 1 : 0);
					outs += ((dot(frustumPlanes[i], vec4(boxMin.x, boxMax.y, boxMax.z, 1.0)) < 0.0) ? 1 : 0);
					outs += ((dot(frustumPlanes[i], vec4(boxMax.x, boxMax.y, boxMax.z, 1.0)) < 0.0) ? 1 : 0);

					if (outs == 8)
					{
						return false;
					}
				}

				// This is not entirely correct (see https://www.iquilezles.org/www/articles/frustumcorrect/frustumcorrect.htm), but good enough.

				return true;
			}

			void main()
			{
				uint threadID = gl_GlobalInvocationID.x;

				if (threadID >= numInputTrianglesBuffer)
				{
					return;
				}

				uvec2 triangleEncoding = inputSubdivisionBuffer[threadID];

				uint rootTriangleID = inputSubdivisionBuffer[threadID].x;
				ivec3 rootTriangle = indexBuffer_fetch(rootTriangleID);
				vec3 rootTriangleVertices[3] = vec3[3](
					vertexBuffer_fetch(rootTriangle.x),
					vertexBuffer_fetch(rootTriangle.y),
					vertexBuffer_fetch(rootTriangle.z));

				uint triangleSubdivisionKey = triangleEncoding.y;

				vec3 triangleVertices[3];
				vec3 parentTriangleVertices[3];
				subd(triangleSubdivisionKey, rootTriangleVertices, triangleVertices, parentTriangleVertices);

				int targetLoD = int(computeLod(triangleVertices));
				int parentLoD = int(computeLod(parentTriangleVertices));

				bool change = updateSubdBuffer(rootTriangleID, triangleSubdivisionKey, targetLoD, parentLoD);

				vec3 bmin = min(min(triangleVertices[0], triangleVertices[1]), triangleVertices[2]);
				bmin.z = 0;

				vec3 bmax = max(max(triangleVertices[0], triangleVertices[1]), triangleVertices[2]);
				bmax.z = heightfieldScaleFactor;

				if (inFrustum(bmin.xyz, bmax.xyz))
				{
					uint nextFreeTriangleIndex = atomicAdd(indirectDrawBuffer.primCount, 1u);

					culledTrianglesBuffer[nextFreeTriangleIndex] = triangleEncoding;

					if (change)
					{
						atomicAdd(changesAtomicCounterBuffer, 1);
					}
				}
			}
		)
	};

	const std::vector<const char*> IndirectCompute
	{
		CommonDefines,
		MULTILINE(
			layout(local_size_x = 1, local_size_y = 1, local_size_z = 1) in;

			layout(std430) restrict writeonly buffer indirectDispatchBufferBlock
			{
				DispatchIndirectCommand indirectDispatchBuffer;
			};

			layout(std430) restrict buffer indirectDrawBufferBlock
			{
				DrawElementsIndirectCommand indirectDrawBuffer;
			};

			layout(std430) restrict writeonly buffer numInputTrianglesBufferBlock
			{
				uint numInputTrianglesBuffer;
			};

			layout(std430) restrict buffer numOutputTrianglesBufferBlock
			{
				uint numOutputTrianglesBuffer;
			};

			void main()
			{
				indirectDispatchBuffer.num_groups_x = (numOutputTrianglesBuffer >> 5u) + 1u;

				numInputTrianglesBuffer = numOutputTrianglesBuffer;

				atomicExchange(numOutputTrianglesBuffer, 0u);
				atomicExchange(indirectDrawBuffer.primCount, 0u);
			}
		)
	};
}

#endif