#pragma once
#ifndef ShaderCodesKhouryEtAl2019_h
#define ShaderCodesKhouryEtAl2019_h

namespace ShaderCodesKhouryEtAl2019
{
#include "../CommonShaders.h"

	const char* CommonDefines = MULTILINE(
		#version 450\n
		#extension GL_EXT_gpu_shader4 : require\n

		const uint bufferCapacity = 1u << 24u; // Note: This value is also defined in the cpp

		// This struct is defined by OpenGL
	struct DispatchIndirectCommand
	{
		uint num_groups_x;
		uint num_groups_y;
		uint num_groups_z;
	};

	// This struct is defined by OpenGL
	struct DrawElementsIndirectCommand
	{
		uint count;
		uint primCount;
		uint firstIndex;
		uint baseVertex;
		uint baseInstance;
	};
	);

	//=================================================================================================
	/**
	*	This code is based on isubd.glsl - public domain implicit subdivision on the GPU (created by
	*	Jonathan Dupuy). See Adaptive GPU Tessellation with Compute Shaders by Jad Khoury, Jonathan
	*	Dupuy, and Christophe Riccio. Adaptive GPU tessellation with compute shaders. In GPU Zen 2:
	*	Advanced Rendering Techniques, W. Engel, Ed. Black Cat Publishing, 2019, ch. 1, pp. 316.
	*/
	//=================================================================================================
	const char* ISUBD = MULTILINE(
		uint parentKey(in uint key)
		{
			return (key >> 1u);
		}

		void childrenKeys(in uint key, out uint children[2])
		{
			children[0] = (key << 1u) | 0u;
			children[1] = (key << 1u) | 1u;
		}

		bool isRootKey(in uint key)
		{
			return (key == 1u);
		}

		bool isLeafKey(in uint key)
		{
			return findMSB(key) == 31;
		}

		bool isChildZeroKey(in uint key)
		{
			return ((key & 1u) == 0u);
		}

		// get xform from bit value
		mat3 bitToXform(in uint bit)
		{
			float s = float(bit) - 0.5;
			vec3 c1 = vec3(s, -0.5, 0);
			vec3 c2 = vec3(-0.5, -s, 0);
			vec3 c3 = vec3(+0.5, +0.5, 1);

			return mat3(c1, c2, c3);
		}

		// get xform from key
		mat3 keyToXform(in uint key)
		{
			mat3 xf = mat3(1.0);

			while (key > 1u)
			{
				xf = bitToXform(key & 1u) * xf;
				key = key >> 1u;
			}

			return xf;
		}

		// get xform from key as well as xform from parent key
		mat3 keyToXform(in uint key, out mat3 xfp)
		{
			// TODO: optimize ?
			xfp = keyToXform(parentKey(key));
			return keyToXform(key);
		}

		// barycentric interpolation
		vec3 berp(in vec3 v[3], in vec2 u)
		{
			return v[0] + u.x * (v[1] - v[0]) + u.y * (v[2] - v[0]);
		}
		vec4 berp(in vec4 v[3], in vec2 u)
		{
			return v[0] + u.x * (v[1] - v[0]) + u.y * (v[2] - v[0]);
		}

		// subdivision routine (vertex position only)
		void subd(in uint key, in vec4 v_in[3], out vec4 v_out[3])
		{
			mat3 xf = keyToXform(key);
			vec2 u1 = (xf * vec3(0, 0, 1)).xy;
			vec2 u2 = (xf * vec3(1, 0, 1)).xy;
			vec2 u3 = (xf * vec3(0, 1, 1)).xy;

			v_out[0] = berp(v_in, u1);
			v_out[1] = berp(v_in, u2);
			v_out[2] = berp(v_in, u3);
		}

		void subd(in uint key, in vec3 v_in[3], out vec3 v_out[3])
		{
			mat3 xf = keyToXform(key);
			vec2 u1 = (xf * vec3(0, 0, 1)).xy;
			vec2 u2 = (xf * vec3(1, 0, 1)).xy;
			vec2 u3 = (xf * vec3(0, 1, 1)).xy;

			v_out[0] = berp(v_in, u1);
			v_out[1] = berp(v_in, u2);
			v_out[2] = berp(v_in, u3);
		}

		// subdivision routine (vertex position only)
		// also computes parent position
		void subd(in uint key, in vec4 v_in[3], out vec4 v_out[3], out vec4 v_out_p[3])
		{
			mat3 xfp; mat3 xf = keyToXform(key, xfp);
			vec2 u1 = (xf * vec3(0, 0, 1)).xy;
			vec2 u2 = (xf * vec3(1, 0, 1)).xy;
			vec2 u3 = (xf * vec3(0, 1, 1)).xy;
			vec2 u4 = (xfp * vec3(0, 0, 1)).xy;
			vec2 u5 = (xfp * vec3(1, 0, 1)).xy;
			vec2 u6 = (xfp * vec3(0, 1, 1)).xy;

			v_out[0] = berp(v_in, u1);
			v_out[1] = berp(v_in, u2);
			v_out[2] = berp(v_in, u3);

			v_out_p[0] = berp(v_in, u4);
			v_out_p[1] = berp(v_in, u5);
			v_out_p[2] = berp(v_in, u6);
		}
		void subd(in uint key, in vec3 v_in[3], out vec3 v_out[3], out vec3 v_out_p[3])
		{
			mat3 xfp; mat3 xf = keyToXform(key, xfp);
			vec2 u1 = (xf * vec3(0, 0, 1)).xy;
			vec2 u2 = (xf * vec3(1, 0, 1)).xy;
			vec2 u3 = (xf * vec3(0, 1, 1)).xy;
			vec2 u4 = (xfp * vec3(0, 0, 1)).xy;
			vec2 u5 = (xfp * vec3(1, 0, 1)).xy;
			vec2 u6 = (xfp * vec3(0, 1, 1)).xy;

			v_out[0] = berp(v_in, u1);
			v_out[1] = berp(v_in, u2);
			v_out[2] = berp(v_in, u3);

			v_out_p[0] = berp(v_in, u4);
			v_out_p[1] = berp(v_in, u5);
			v_out_p[2] = berp(v_in, u6);
		}

		layout(std430) restrict writeonly buffer outputSubdivisionBufferBlock
		{
			uvec2 outputSubdivisionBuffer[];
		};

		layout(std430) restrict buffer numOutputTrianglesBufferBlock
		{
			uint numOutputTrianglesBuffer;
		};

		void writeKey(uint primID, uint key)
		{
			uint idx = atomicAdd(numOutputTrianglesBuffer, 1u);

			if (idx < bufferCapacity)
			{
				outputSubdivisionBuffer[idx] = uvec2(primID, key);
			}
			else
			{
				// Note: We discard all triangles that do not fit in the output buffer.
				// This will inevitably break the triangulation. A proper mechanism is
				// needed to guarantee that parent triangles will be stored in case a
				// child triangle cannot be written anymore.

				const uint lastIndex = bufferCapacity - 1u;

				atomicExchange(numOutputTrianglesBuffer, lastIndex);
			}
		}

		bool updateSubdBuffer(uint primID, uint key, int targetLod, int parentLod)
		{
			int keyLod = findMSB(key);

			if (keyLod < targetLod && !isLeafKey(key))
			{
				uint children[2];
				childrenKeys(key, children);

				writeKey(primID, children[0]);
				writeKey(primID, children[1]);

				return true;
			}
			else if (keyLod < (parentLod + 1))
			{
				writeKey(primID, key);
			}
			else
			{
				if (isRootKey(key))
				{
					writeKey(primID, key);
				}
				else if (isChildZeroKey(key))
				{
					writeKey(primID, parentKey(key));

					return true;
				}
			}

			return false;
		}

		uniform float lodFactor;

		float distanceToLod(float z, float lodFactor)
		{
			float lod = -2.0 * log2(clamp(z * lodFactor, 0.0, 1.0));

			return min(lod, 30.0);
		}

		float computeLod(vec3 c)
		{
			// Heightfield sampling, which is just a noise function here
			c.z += sampleHeightfield(inverseViewMatrix[3].xy);

			vec3 cxf = (viewMatrix * vec4(c, 1.0)).xyz;
			float z = length(cxf);

			return distanceToLod(z, lodFactor);
		}

		float computeLod(in vec3 v[3])
		{
			vec3 c = (v[1].xyz + v[2].xyz) * 0.5;

			return computeLod(c);
		}
	);

	const std::vector<const char*> TerrainVertex
	{
		CommonDefines,
		MULTILINE(
			in vec2 inPosition;
		),
		CameraUniforms,
		HeightfieldSampling,
		ISUBD,
		MULTILINE(
			layout(std430) restrict readonly buffer vertexBufferBlock
			{
				float vertexBuffer[];
			};

			vec3 vertexBuffer_fetch(int elementIndex)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				return vec3(vertexBuffer[elementOffsets.x], vertexBuffer[elementOffsets.y], vertexBuffer[elementOffsets.z]);
			}

			layout(std430) restrict readonly buffer indexBufferBlock
			{
				int indexBuffer[];
			};

			ivec3 indexBuffer_fetch(uint elementIndex)
			{
				uvec3 elementOffsets = uvec3(3u * elementIndex) + uvec3(0u, 1u, 2u);

				return ivec3(indexBuffer[elementOffsets.x], indexBuffer[elementOffsets.y], indexBuffer[elementOffsets.z]);
			}

			layout(std430) restrict readonly buffer culledTrianglesBufferBlock
			{
				uvec2 culledTrianglesBuffer[];
			};

			void main()
			{
				int threadID = gl_InstanceID;

				uvec2 triangleEncoding = culledTrianglesBuffer[threadID];

				uint rootTriangleID = triangleEncoding.x;
				ivec3 rootTriangle = indexBuffer_fetch(rootTriangleID);
				vec3 rootTriangleVertices[3] = vec3[3](
					vertexBuffer_fetch(rootTriangle.x),
					vertexBuffer_fetch(rootTriangle.y),
					vertexBuffer_fetch(rootTriangle.z));

				uint triangleSubdivisionKey = triangleEncoding.y;
				vec3 triangleVertices[3];
				subd(triangleSubdivisionKey, rootTriangleVertices, triangleVertices);

				vec3 finalVertex = berp(triangleVertices, inPosition);

				// Heightfield sampling, which is just a noise function here
				finalVertex.z = sampleHeightfield(finalVertex.xy);

				gl_Position = viewProjectionMatrix * vec4(finalVertex, 1.0);
			}
		)
	};

	const std::vector<const char*> LoDCompute
	{
		CommonDefines,
		CameraUniforms,
		HeightfieldSampling,
		ISUBD,
		MULTILINE(
			layout(local_size_x = 32, local_size_y = 1, local_size_z = 1) in;

			layout(std430) restrict readonly buffer vertexBufferBlock
			{
				float vertexBuffer[];
			};

			vec3 vertexBuffer_fetch(int elementIndex)
			{
				ivec3 elementOffsets = ivec3(3 * elementIndex) + ivec3(0, 1, 2);

				return vec3(vertexBuffer[elementOffsets.x], vertexBuffer[elementOffsets.y], vertexBuffer[elementOffsets.z]);
			}

			layout(std430) restrict readonly buffer indexBufferBlock
			{
				int indexBuffer[];
			};

			ivec3 indexBuffer_fetch(uint elementIndex)
			{
				uvec3 elementOffsets = uvec3(3u * elementIndex) + uvec3(0u, 1u, 2u);

				return ivec3(indexBuffer[elementOffsets.x], indexBuffer[elementOffsets.y], indexBuffer[elementOffsets.z]);
			}

			layout(std430) restrict readonly buffer inputSubdivisionBufferBlock
			{
				uvec2 inputSubdivisionBuffer[];
			};

			layout(std430) restrict writeonly buffer culledTrianglesBufferBlock
			{
				uvec2 culledTrianglesBuffer[];
			};

			layout(std430) restrict readonly buffer numInputTrianglesBufferBlock
			{
				uint numInputTrianglesBuffer;
			};

			layout(std430) restrict buffer indirectDrawBufferBlock
			{
				DrawElementsIndirectCommand indirectDrawBuffer;
			};

			layout(std430) restrict buffer changesAtomicCounterBufferBlock
			{
				int changesAtomicCounterBuffer;
			};

			bool inFrustum(vec3 boxMin, vec3 boxMax)
			{
				vec4 frustumPlanes[6] = vec4[](
					frustumPlaneLeft,
					frustumPlaneRight,
					frustumPlaneBottom,
					frustumPlaneTop,
					frustumPlaneFar,
					frustumPlaneNear);

				for (int i = 0; i < 6; i++)
				{
					int outs = 0;

					outs += ((dot(frustumPlanes[i], vec4(boxMin.x, boxMin.y, boxMin.z, 1.0)) < 0.0) ? 1 : 0);
					outs += ((dot(frustumPlanes[i], vec4(boxMax.x, boxMin.y, boxMin.z, 1.0)) < 0.0) ? 1 : 0);
					outs += ((dot(frustumPlanes[i], vec4(boxMin.x, boxMax.y, boxMin.z, 1.0)) < 0.0) ? 1 : 0);
					outs += ((dot(frustumPlanes[i], vec4(boxMax.x, boxMax.y, boxMin.z, 1.0)) < 0.0) ? 1 : 0);
					outs += ((dot(frustumPlanes[i], vec4(boxMin.x, boxMin.y, boxMax.z, 1.0)) < 0.0) ? 1 : 0);
					outs += ((dot(frustumPlanes[i], vec4(boxMax.x, boxMin.y, boxMax.z, 1.0)) < 0.0) ? 1 : 0);
					outs += ((dot(frustumPlanes[i], vec4(boxMin.x, boxMax.y, boxMax.z, 1.0)) < 0.0) ? 1 : 0);
					outs += ((dot(frustumPlanes[i], vec4(boxMax.x, boxMax.y, boxMax.z, 1.0)) < 0.0) ? 1 : 0);

					if (outs == 8)
					{
						return false;
					}
				}

				// This is not entirely correct (see https://www.iquilezles.org/www/articles/frustumcorrect/frustumcorrect.htm), but good enough.

				return true;
			}

			void main()
			{
				uint threadID = gl_GlobalInvocationID.x;

				if (threadID >= numInputTrianglesBuffer)
				{
					return;
				}

				uvec2 triangleEncoding = inputSubdivisionBuffer[threadID];

				uint rootTriangleID = inputSubdivisionBuffer[threadID].x;
				ivec3 rootTriangle = indexBuffer_fetch(rootTriangleID);
				vec3 rootTriangleVertices[3] = vec3[3](
					vertexBuffer_fetch(rootTriangle.x),
					vertexBuffer_fetch(rootTriangle.y),
					vertexBuffer_fetch(rootTriangle.z));

				uint triangleSubdivisionKey = triangleEncoding.y;

				vec3 triangleVertices[3];
				vec3 parentTriangleVertices[3];
				subd(triangleSubdivisionKey, rootTriangleVertices, triangleVertices, parentTriangleVertices);

				int targetLoD = int(computeLod(triangleVertices));
				int parentLoD = int(computeLod(parentTriangleVertices));

				bool change = updateSubdBuffer(rootTriangleID, triangleSubdivisionKey, targetLoD, parentLoD);

				vec3 bmin = min(min(triangleVertices[0], triangleVertices[1]), triangleVertices[2]);
				bmin.z = 0;

				vec3 bmax = max(max(triangleVertices[0], triangleVertices[1]), triangleVertices[2]);
				bmax.z = heightfieldScaleFactor;

				if (inFrustum(bmin.xyz, bmax.xyz))
				{
					uint nextFreeTriangleIndex = atomicAdd(indirectDrawBuffer.primCount, 1u);

					culledTrianglesBuffer[nextFreeTriangleIndex] = triangleEncoding;

					if (change)
					{
						atomicAdd(changesAtomicCounterBuffer, 1);
					}
				}
			}
		)
	};

	const std::vector<const char*> IndirectCompute
	{
		CommonDefines,
		MULTILINE(
			layout(local_size_x = 1, local_size_y = 1, local_size_z = 1) in;

			layout(std430) restrict writeonly buffer indirectDispatchBufferBlock
			{
				DispatchIndirectCommand indirectDispatchBuffer;
			};

			layout(std430) restrict buffer indirectDrawBufferBlock
			{
				DrawElementsIndirectCommand indirectDrawBuffer;
			};

			layout(std430) restrict writeonly buffer numInputTrianglesBufferBlock
			{
				uint numInputTrianglesBuffer;
			};

			layout(std430) restrict buffer numOutputTrianglesBufferBlock
			{
				uint numOutputTrianglesBuffer;
			};

			void main()
			{
				indirectDispatchBuffer.num_groups_x = (numOutputTrianglesBuffer >> 5u) + 1u;

				numInputTrianglesBuffer = numOutputTrianglesBuffer;

				atomicExchange(numOutputTrianglesBuffer, 0u);
				atomicExchange(indirectDrawBuffer.primCount, 0u);
			}
		)
	};
}

#endif