#pragma once
#ifndef ShaderCodesLeeEtAl2013_h
#define ShaderCodesLeeEtAl2013_h

namespace ShaderCodesLeeEtAl2013
{
	#include "../CommonShaders.h"

	const char* CommonLeeEtAl2013 = MULTILINE(
		const float maximumEdgeLengthInPixels = 10.0;
		const float minimumEdgeLengthInMeters = 1.0;

		float getEdgeLengthInPixels(vec4 worldSpacePosition0, vec4 worldSpacePosition1)
		{
			float worldSpaceEdgeLength = distance(worldSpacePosition0, worldSpacePosition1);
			vec4 worldSpaceCenterPosition = 0.5 * (worldSpacePosition0 + worldSpacePosition0);

			vec4 viewSpaceCenterPosition = viewMatrix * worldSpaceCenterPosition;
			vec4 viewSpaceOffsetPosition = viewSpaceCenterPosition + vec4(worldSpaceEdgeLength, 0.0, 0.0, 0.0);

			vec4 clipSpaceCenterPosition = projectionMatrix * viewSpaceCenterPosition;
			vec4 clipSpaceOffsetPosition = projectionMatrix * viewSpaceOffsetPosition;

			vec2 pixelSpaceCenterPosition = (0.5 * clipSpaceCenterPosition.xy / clipSpaceCenterPosition.w + 0.5) * viewPortSize;
			vec2 pixelSpaceOffsetPosition = (0.5 * clipSpaceOffsetPosition.xy / clipSpaceOffsetPosition.w + 0.5) * viewPortSize;

			return distance(pixelSpaceCenterPosition, pixelSpaceOffsetPosition);
		}
	);

	const std::vector<const char*> VertexShader
	{
		MULTILINE(
			#version 450\n

			in vec4 inPosition;
		),
		HeightfieldSampling,
		MULTILINE(
			void main()
			{
				gl_Position = vec4(inPosition.xy, sampleHeightfield(inPosition.xy), 1.0);
			}
		)
	};

	const std::vector<const char*> ControlShader
	{
		MULTILINE(
			#version 450\n

			layout(vertices = 3) out;

			uniform vec2 heightfieldCellSize;
		),
		CameraUniforms,
		CommonLeeEtAl2013,
		MULTILINE(
			bool inFrustum(vec4 worldSpacePosition, float maxDistance)
			{
				return !((dot(worldSpacePosition, frustumPlaneLeft) <= maxDistance) ||
					(dot(worldSpacePosition, frustumPlaneRight) <= maxDistance) ||
					(dot(worldSpacePosition, frustumPlaneBottom) <= maxDistance) ||
					(dot(worldSpacePosition, frustumPlaneTop) <= maxDistance) ||
					(dot(worldSpacePosition, frustumPlaneFar) <= maxDistance) ||
					(dot(worldSpacePosition, frustumPlaneNear) <= maxDistance));
			}

			void main()
			{
				gl_out[gl_InvocationID].gl_Position = gl_in[gl_InvocationID].gl_Position;

				gl_TessLevelOuter[0] = 0.0;
				gl_TessLevelOuter[1] = 0.0;
				gl_TessLevelOuter[2] = 0.0;
				gl_TessLevelOuter[3] = 0.0;

				gl_TessLevelInner[0] = 0.0;
				gl_TessLevelInner[1] = 0.0;

				if (gl_InvocationID > 0)
				{
					return;
				}
				else
				{
					vec3 worldSpaceEdgeLengths = vec3(
						distance(gl_in[1].gl_Position, gl_in[2].gl_Position),
						distance(gl_in[2].gl_Position, gl_in[0].gl_Position),
						distance(gl_in[0].gl_Position, gl_in[1].gl_Position));

					float offset = -max(worldSpaceEdgeLengths.x, max(worldSpaceEdgeLengths.y, worldSpaceEdgeLengths.z));

					if ((inFrustum(gl_in[0].gl_Position, offset)) || (inFrustum(gl_in[1].gl_Position, offset)) || (inFrustum(gl_in[2].gl_Position, offset)))
					{
						vec3 worldSpaceEdgeFactors = worldSpaceEdgeLengths / minimumEdgeLengthInMeters;

						vec3 pixelSpaceEdgeLengths = vec3(
							getEdgeLengthInPixels(gl_in[1].gl_Position, gl_in[2].gl_Position),
							getEdgeLengthInPixels(gl_in[2].gl_Position, gl_in[0].gl_Position),
							getEdgeLengthInPixels(gl_in[0].gl_Position, gl_in[1].gl_Position));

						vec3 pixelSpaceEdgeFactors = pixelSpaceEdgeLengths / maximumEdgeLengthInPixels;

						vec3 tessellationFactors = clamp(min(worldSpaceEdgeFactors, pixelSpaceEdgeFactors), 1.0, 64.0);

						gl_TessLevelOuter[0] = tessellationFactors.x;
						gl_TessLevelOuter[1] = tessellationFactors.y;
						gl_TessLevelOuter[2] = tessellationFactors.z;
						gl_TessLevelOuter[3] = 0.0;

						gl_TessLevelInner[0] = min(tessellationFactors.x, min(tessellationFactors.y, tessellationFactors.z));
						gl_TessLevelInner[1] = 0.0;
					}
				}
			}
		)
	};

	const std::vector<const char*> EvaluationShader
	{
		MULTILINE(
			#version 450\n

			layout(triangles, fractional_odd_spacing, ccw) in;

			void main()
			{
				gl_Position = gl_TessCoord.x * gl_in[0].gl_Position + gl_TessCoord.y * gl_in[1].gl_Position + gl_TessCoord.z * gl_in[2].gl_Position;
			}
		)
	};

	const std::vector<const char*> GeometryShader
	{
		MULTILINE(
			#version 450\n

			layout(triangles) in;
			layout(triangle_strip, max_vertices = 3) out;

			out vec3 gsHeights;

			layout(std430) restrict buffer atomicCounterBlock
			{
				int atomicCounter;
			};

			layout(std430) restrict writeonly buffer outputVerticesBlock
			{
				vec4 outputVertices[];
			};

			uniform int maxVertices;
			uniform bool abort;
		),
		HeightfieldSampling,
		CameraUniforms,
		CommonLeeEtAl2013,
		MULTILINE(
			void main()
			{
				vec4 worldSpacePosition0 = vec4(gl_in[0].gl_Position.xy, sampleHeightfield(gl_in[0].gl_Position.xy), 1.0);
				vec4 worldSpacePosition1 = vec4(gl_in[1].gl_Position.xy, sampleHeightfield(gl_in[1].gl_Position.xy), 1.0);
				vec4 worldSpacePosition2 = vec4(gl_in[2].gl_Position.xy, sampleHeightfield(gl_in[2].gl_Position.xy), 1.0);

				vec3 worldSpaceEdgeLengths = vec3(
					distance(gl_in[1].gl_Position, gl_in[2].gl_Position),
					distance(gl_in[2].gl_Position, gl_in[0].gl_Position),
					distance(gl_in[0].gl_Position, gl_in[1].gl_Position));

				vec3 pixelSpaceEdgeLengths = vec3(
					getEdgeLengthInPixels(worldSpacePosition1, worldSpacePosition2),
					getEdgeLengthInPixels(worldSpacePosition2, worldSpacePosition0),
					getEdgeLengthInPixels(worldSpacePosition0, worldSpacePosition1));

				bvec3 worldSpaceRequirement = greaterThan(worldSpaceEdgeLengths, vec3(minimumEdgeLengthInMeters));
				bvec3 pixelSpaceRequirement = greaterThan(pixelSpaceEdgeLengths, vec3(maximumEdgeLengthInPixels));

				if ((!abort) && (any((worldSpaceRequirement) && (pixelSpaceRequirement))))
				{
					int nextFreeVertexIndex = atomicAdd(atomicCounter, 3);

					if (nextFreeVertexIndex + 2 < maxVertices)
					{
						outputVertices[nextFreeVertexIndex] = worldSpacePosition0;
						outputVertices[nextFreeVertexIndex + 1] = worldSpacePosition1;
						outputVertices[nextFreeVertexIndex + 2] = worldSpacePosition2;

						return;
					}
				}

				const vec3[3] heights = vec3[3](vec3(1.0, 0.0, 0.0), vec3(0.0, 1.0, 0.0), vec3(0.0, 0.0, 1.0));

				gsHeights = heights[0];
				gl_Position = viewProjectionMatrix * worldSpacePosition0;
				EmitVertex();

				gsHeights = heights[1];
				gl_Position = viewProjectionMatrix * worldSpacePosition1;
				EmitVertex();

				gsHeights = heights[2];
				gl_Position = viewProjectionMatrix * worldSpacePosition2;
				EmitVertex();

				EndPrimitive();
			}
		)
	};
}

#endif