#include "TessellationLeeEtAl2013.h"

#include <cmath>
#include <ppl.h>
#include <vector>

#include <GL/glew.h>

#include "../Heightfield.h"
#include "ShaderCodesLeeEtAl2013.h"

#include "../PipelineStatisticsQuery.h"

using namespace Eigen;

const int maxInterations = 3;
const int maxVertices = 3 << 23;

//=================================================================================================
/**
*	Constructor.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
TessellationLeeEtAl2013::TessellationLeeEtAl2013(const Heightfield& heightfield) : AbstractTessellation(heightfield, "LeeEtAl2013")
{
	const ChunkDescriptor& chunkDescriptor = mHeightfield.mChunkDescriptor;

	//=================================================================================================
	/**
	*	Create rendering program
	*/
	//=================================================================================================

	mProgram = std::make_unique<Program>(
		ShaderCodesLeeEtAl2013::VertexShader,
		ShaderCodesLeeEtAl2013::ControlShader,
		ShaderCodesLeeEtAl2013::EvaluationShader,
		ShaderCodesLeeEtAl2013::GeometryShader,
		ShaderCodesLeeEtAl2013::CommonFragmentShader);
	mProgram->setUniform2i("heightfieldDimensions", chunkDescriptor.mChunkGridDimensions);
	mProgram->setUniform2f("heightfieldOrigin", mHeightfield.mOrigin);
	mProgram->setUniform2f("heightfieldCellSize", chunkDescriptor.mChunkCellSize);
	mProgram->setUniform2f("inverseHeightfieldCellSize", Vector2f::Ones().cwiseQuotient(chunkDescriptor.mChunkCellSize));
	mProgram->setUniform1f("heightfieldScaleFactor", mHeightfield.mScaleFactor);
	mProgram->setUniform2i("gridDimensions", mHeightfield.mDimensions);
	mProgram->setUniform2f("domainOrigin", mHeightfield.mOrigin);
	mProgram->setUniform2f("cellSize", mHeightfield.mCellSize);

	//=================================================================================================
	/**
	*	Create vertex buffers
	*/
	//=================================================================================================

	std::vector<Vector4f> vertices(6 * chunkDescriptor.mChunkGridDimensions.x() * chunkDescriptor.mChunkGridDimensions.y());

	Concurrency::parallel_for(0, chunkDescriptor.mChunkGridDimensions.x() * chunkDescriptor.mChunkGridDimensions.y(), [&](int cellIndex)
	{
		int x = cellIndex % chunkDescriptor.mChunkGridDimensions.x();
		int y = cellIndex / chunkDescriptor.mChunkGridDimensions.x();

		Vector4f bottomLeftVertex(
			mHeightfield.mOrigin.x() + static_cast<float>(x) * chunkDescriptor.mChunkCellSize.x(),
			mHeightfield.mOrigin.y() + static_cast<float>(y) * chunkDescriptor.mChunkCellSize.y(),
			0.0f,
			1.0f);
		Vector4f bottomRightVertex(
			mHeightfield.mOrigin.x() + static_cast<float>(x + 1) * chunkDescriptor.mChunkCellSize.x(),
			mHeightfield.mOrigin.y() + static_cast<float>(y) * chunkDescriptor.mChunkCellSize.y(),
			0.0f,
			1.0f);
		Vector4f topLeftVertex(
			mHeightfield.mOrigin.x() + static_cast<float>(x) * chunkDescriptor.mChunkCellSize.x(),
			mHeightfield.mOrigin.y() + static_cast<float>(y + 1) * chunkDescriptor.mChunkCellSize.y(),
			0.0f,
			1.0f);
		Vector4f topRightVertex(
			mHeightfield.mOrigin.x() + static_cast<float>(x + 1) * chunkDescriptor.mChunkCellSize.x(),
			mHeightfield.mOrigin.y() + static_cast<float>(y + 1) * chunkDescriptor.mChunkCellSize.y(),
			0.0f,
			1.0f);

		int bottomLeftVertexIndex = y * (chunkDescriptor.mChunkGridDimensions.x() + 1) + x;
		int bottomRightVertexIndex = bottomLeftVertexIndex + 1;
		int topLeftVertexIndex = (y + 1) * (chunkDescriptor.mChunkGridDimensions.x() + 1) + x;
		int topRightVertexIndex = topLeftVertexIndex + 1;

		vertices[6 * cellIndex] = bottomLeftVertex;
		vertices[6 * cellIndex + 1] = bottomRightVertex;
		vertices[6 * cellIndex + 2] = topLeftVertex;

		vertices[6 * cellIndex + 3] = topLeftVertex;
		vertices[6 * cellIndex + 4] = bottomRightVertex;
		vertices[6 * cellIndex + 5] = topRightVertex;
	});

	glCreateBuffers(1, &mChunkVertexBuffer);
	glNamedBufferData(mChunkVertexBuffer, sizeof(Vector4f) * vertices.size(), vertices.data(), GL_STATIC_DRAW);

	glCreateBuffers(1, &mVertexBuffer0);
	glNamedBufferData(mVertexBuffer0, sizeof(Vector4f) * maxVertices, nullptr, GL_STATIC_DRAW);

	glCreateBuffers(1, &mVertexBuffer1);
	glNamedBufferData(mVertexBuffer1, sizeof(Vector4f) * maxVertices, nullptr, GL_STATIC_DRAW);

	//=================================================================================================
	/**
	*	Create atomic counter
	*/
	//=================================================================================================

	glCreateBuffers(1, &mAtomicCounter);
	glNamedBufferData(mAtomicCounter, sizeof(GLint), nullptr, GL_STATIC_DRAW);

	//=================================================================================================
	/**
	*	Create vertex array object
	*/
	//=================================================================================================

	glCreateVertexArrays(1, &mVAO);

	GLuint positionLocation = mProgram->getAttributeLocation("inPosition");
	glEnableVertexArrayAttrib(mVAO, positionLocation);
	glVertexArrayAttribFormat(mVAO, positionLocation, 4, GL_FLOAT, GL_FALSE, 0);
	glVertexArrayVertexBuffer(mVAO, positionLocation, mChunkVertexBuffer, 0, sizeof(Vector4f));

	glPatchParameteri(GL_PATCH_VERTICES, 3);
}

//=================================================================================================
/**
*	Destructor.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
TessellationLeeEtAl2013::~TessellationLeeEtAl2013()
{
	glDeleteVertexArrays(1, &mVAO);

	glDeleteBuffers(1, &mChunkVertexBuffer);
	glDeleteBuffers(1, &mVertexBuffer0);
	glDeleteBuffers(1, &mVertexBuffer1);
	glDeleteBuffers(1, &mAtomicCounter);
}

//=================================================================================================
/**
*	Performs a per-frame update of the tessellation.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void TessellationLeeEtAl2013::update(const Camera& camera)
{

}

//=================================================================================================
/**
*	Renders the geometry.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void TessellationLeeEtAl2013::render(const Camera& camera)
{
	GLint numInputVertices = 6 * mHeightfield.mChunkDescriptor.mChunkGridDimensions.x() * mHeightfield.mChunkDescriptor.mChunkGridDimensions.y();
	GLuint inputVertexBuffer = mVertexBuffer0;
	GLuint outputVertexBuffer = mVertexBuffer1;

	mProgram->bind();
	mProgram->setCamera(camera);
	mProgram->setUniform1i("maxVertices", maxVertices);
	mProgram->setSSBO("atomicCounter", mAtomicCounter);
	mProgram->setSSBO("heightfieldBuffer", mHeightfield.mBuffer);

	glBindVertexArray(mVAO);

	for (int iteration = 0; iteration < maxInterations; iteration++)
	{
		mProgram->setUniformBool("abort", (iteration == maxInterations - 1));
		mProgram->setSSBO("outputVertices", outputVertexBuffer);

		glClearNamedBufferData(mAtomicCounter, GL_R32I, GL_RED_INTEGER, GL_INT, nullptr);

		glVertexArrayVertexBuffer(mVAO, mProgram->getAttributeLocation("inPosition"), (iteration == 0) ? mChunkVertexBuffer : inputVertexBuffer, 0, sizeof(Vector4f));

		glDrawArrays(GL_PATCHES, 0, numInputVertices);

		glMemoryBarrier(GL_BUFFER_UPDATE_BARRIER_BIT | GL_SHADER_STORAGE_BARRIER_BIT);

		GLint numOutputVertices = 0;
		glGetNamedBufferSubData(mAtomicCounter, 0, sizeof(GLint), &numOutputVertices);

		GLuint temp = inputVertexBuffer;
		inputVertexBuffer = outputVertexBuffer;
		outputVertexBuffer = temp;

		numInputVertices = numOutputVertices;

		if (numInputVertices == 0u)
		{
			break;
		}
	}

	glBindBuffer(GL_SHADER_STORAGE_BUFFER, 0u);
	glBindVertexArray(0u);

	mProgram->unbind();
}