#include "TessellationNaive.h"

//#include <cmath>
#include <ppl.h>
#include <vector>

#include <GL/glew.h>

#include "../Heightfield.h"
#include "ShaderCodesNaive.h"

#include "../PipelineStatisticsQuery.h"

using namespace Eigen;

//=================================================================================================
/**
*	Constructor.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
TessellationNaive::TessellationNaive(const Heightfield& heightfield) : AbstractTessellation(heightfield, "Naive")
{
	const ChunkDescriptor& chunkDescriptor = mHeightfield.mChunkDescriptor;

	//=================================================================================================
	/**
	*	Create vertex buffer
	*/
	//=================================================================================================

	std::vector<Vector4f> vertices((mHeightfield.mDimensions.x() + 1) * (mHeightfield.mDimensions.y() + 1));

	Concurrency::parallel_for(0, static_cast<int>(vertices.size()), [&](int vertexIndex)
	{
		int x = vertexIndex % (mHeightfield.mDimensions.x() + 1);
		int y = vertexIndex / (mHeightfield.mDimensions.x() + 1);

		vertices[vertexIndex] = Vector4f(
			mHeightfield.mOrigin.x() + static_cast<float>(x) * mHeightfield.mCellSize.x(),
			mHeightfield.mOrigin.y() + static_cast<float>(y) * mHeightfield.mCellSize.y(),
			0.0f,
			1.0f);
	});

	glCreateBuffers(1, &mVertexBuffer);
	glNamedBufferData(mVertexBuffer, sizeof(Vector4f) * vertices.size(), vertices.data(), GL_STATIC_DRAW);

	//=================================================================================================
	/**
	*	Create index buffer
	*/
	//=================================================================================================

	std::vector<Vector3i> indices(2 * mHeightfield.mDimensions.x() * mHeightfield.mDimensions.y());

	Concurrency::parallel_for(0, mHeightfield.mDimensions.x() * mHeightfield.mDimensions.y(), [&](int cellIndex)
	{
		int x = cellIndex % mHeightfield.mDimensions.x();
		int y = cellIndex / mHeightfield.mDimensions.x();

		int bottomLeftVertexIndex = y * (mHeightfield.mDimensions.x() + 1) + x;
		int bottomRightVertexIndex = bottomLeftVertexIndex + 1;
		int topLeftVertexIndex = (y + 1) * (mHeightfield.mDimensions.x() + 1) + x;
		int topRightVertexIndex = topLeftVertexIndex + 1;

		indices[2 * cellIndex] = Vector3i(bottomLeftVertexIndex, bottomRightVertexIndex, topLeftVertexIndex);
		indices[2 * cellIndex + 1] = Vector3i(topRightVertexIndex, topLeftVertexIndex, bottomRightVertexIndex);
	});

	glCreateBuffers(1, &mIndexBuffer);
	glNamedBufferData(mIndexBuffer, sizeof(Vector3i) * indices.size(), indices.data(), GL_STATIC_DRAW);

	//=================================================================================================
	/**
	*	Create rendering program
	*/
	//=================================================================================================

	mProgram = std::make_unique<Program>(
		ShaderCodesNaive::VertexShader,
		ShaderCodesNaive::CommonGeometryShader,
		ShaderCodesNaive::CommonFragmentShader);
	mProgram->setUniform2i("heightfieldDimensions", mHeightfield.mDimensions);
	mProgram->setUniform2f("heightfieldOrigin", mHeightfield.mOrigin);
	mProgram->setUniform2f("heightfieldCellSize", mHeightfield.mCellSize);
	mProgram->setUniform2f("inverseHeightfieldCellSize", Vector2f::Ones().cwiseQuotient(mHeightfield.mCellSize));
	mProgram->setUniform1f("heightfieldScaleFactor", mHeightfield.mScaleFactor);
	mProgram->setUniform2i("gridDimensions", mHeightfield.mDimensions);
	mProgram->setUniform2f("domainOrigin", mHeightfield.mOrigin);
	mProgram->setUniform2f("cellSize", mHeightfield.mCellSize);

	//=================================================================================================
	/**
	*	Create vertex array object
	*/
	//=================================================================================================

	glCreateVertexArrays(1, &mVAO);
	glVertexArrayElementBuffer(mVAO, mIndexBuffer);

	GLint positionAttributeLocation = mProgram->getAttributeLocation("inPosition");
	glEnableVertexArrayAttrib(mVAO, positionAttributeLocation);
	glVertexArrayAttribFormat(mVAO, positionAttributeLocation, 4, GL_FLOAT, GL_FALSE, 0);
	glVertexArrayVertexBuffer(mVAO, positionAttributeLocation, mVertexBuffer, 0, sizeof(Vector4f));
}

//=================================================================================================
/**
*	Destructor.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
TessellationNaive::~TessellationNaive()
{
	glDeleteVertexArrays(1, &mVAO);

	glDeleteBuffers(1, &mVertexBuffer);
	glDeleteBuffers(1, &mIndexBuffer);
}

//=================================================================================================
/**
*	Performs a per-frame update of the tessellation.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void TessellationNaive::update(const Camera& camera)
{

}

//=================================================================================================
/**
*	Renders the geometry.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void TessellationNaive::render(const Camera& camera)
{
	mProgram->bind();
	mProgram->setCamera(camera);
	mProgram->setSSBO("heightfieldBuffer", mHeightfield.mBuffer);

	glBindVertexArray(mVAO);

	glDrawElements(GL_TRIANGLES, 6 * mHeightfield.mDimensions.x() * mHeightfield.mDimensions.y(), GL_UNSIGNED_INT, 0);

	glBindBuffer(GL_SHADER_STORAGE_BUFFER, 0);
	glBindVertexArray(0u);

	mProgram->unbind();
}