#include "PipelineStatisticsQuery.h"

#include <algorithm>
#include <iostream>
#include <map>
#include <sstream>
#include <string>

std::vector<std::pair<GLenum, GLuint>> PipelineStatisticsQuery::sQueryObjects;
std::vector<std::pair<std::string, GLuint>> PipelineStatisticsQuery::sQueryResults;

//=================================================================================================
/**
*	Creates OpenGL query objects for the different statistics and starts the recording.
*
*	@author	Daniel Cornel
*
*	@param	statistics	A list of all statistics to record
**/
//=================================================================================================
void PipelineStatisticsQuery::start(const std::vector<GLenum>& statistics)
{
	for (const std::pair<GLenum, GLuint>& queryObject : sQueryObjects)
	{
		glDeleteQueries(1, &queryObject.second);
	}

	sQueryObjects.clear();

	glFinish();

	for (GLenum pipelineStatistic : statistics)
	{
		GLuint newQuery;
		glGenQueries(1, &newQuery);
		glBeginQuery(pipelineStatistic, newQuery);

		sQueryObjects.emplace_back(std::make_pair(pipelineStatistic, newQuery));
	}
}

//=================================================================================================
/**
*	Creates OpenGL query objects for the different statistics and starts the recording.
*
*	@author	Daniel Cornel
*
*	@param	statistic	The statistic to record
**/
//=================================================================================================
void PipelineStatisticsQuery::start(GLenum statistic)
{
	start(std::vector<GLenum> { statistic });
}

//=================================================================================================
/**
*	Creates OpenGL query objects for the different statistics and starts the recording. Will record
*	all available statistics.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void PipelineStatisticsQuery::start()
{
	start(PipelineStatistics::sPipelineStatistics);
}

//=================================================================================================
/**
*	Stops the recording and destroys the OpenGL query objects for the different statistics. Dumps
*	the results to std::cout.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
bool PipelineStatisticsQuery::stop()
{
	glFinish();

	sQueryResults.clear();

	bool error = false;

	for (auto it = sQueryObjects.rbegin(); it != sQueryObjects.rend(); it++)
	{
		glEndQuery(it->first);

		const std::string& pipelineStatisticsString = PipelineStatistics::sPipelineStatisticsStrings.at(it->first);

		GLuint queryResult;
		glGetQueryObjectuiv(it->second, GL_QUERY_RESULT, &queryResult);

		sQueryResults.push_back(std::make_pair(pipelineStatisticsString, queryResult));

		if ((it->first == GL_TIME_ELAPSED) && (queryResult == 0xFFFFFFFFu))
		{
			error = true;
		}

		glDeleteQueries(1, &it->second);
	}

	sQueryObjects.clear();

	dump(error);

	return error;
}

//=================================================================================================
/**
*	Writes all query results to stdout.
*
*	@author	Daniel Cornel
*
*	@param	error	Flag to indicate that an error occurred
**/
//=================================================================================================
void PipelineStatisticsQuery::dump(bool error)
{
	std::cout << "\nPipeline query statistics:\n";

	for (const std::pair<std::string, GLuint>& queryResult : sQueryResults)
	{
		if (queryResult.first == "Time elapsed")
		{
			std::cout << queryResult.first << ": " << formatTimeString(queryResult.second) << "\n";
		}
		else
		{
			std::cout << queryResult.first << ": " << std::to_string(queryResult.second) << "\n";
		}
	}

	if (error)
	{
		std::cout << "An error has occurred during the pipeline query. The OpenGL context might be corrupted.\n";
	}
}

//=================================================================================================
/**
*	Format the time given in nanoseconds into a string with appropriate units.
*
*	@author	Daniel Cornel
*
*	@param	timeInNS	The time in nanoseconds
**/
//=================================================================================================
std::string PipelineStatisticsQuery::formatTimeString(GLuint timeInNS)
{
	std::stringstream timeSStr;
	bool unitWritten = false;

	GLuint remainingTime = timeInNS;
	GLuint seconds = timeInNS / 1000000000u;

	timeInNS -= seconds * 1000000000u;

	GLuint milliseconds = timeInNS / 1000000u;

	timeInNS -= milliseconds * 1000000u;

	GLuint microseconds = timeInNS / 1000u;
	GLuint nanoseconds = timeInNS - microseconds * 1000u;

	if (seconds > 0u)
	{
		const std::string str = std::to_string(seconds);
		timeSStr << str << " s";

		unitWritten = true;
	}

	if (milliseconds > 0u)
	{
		if (unitWritten)
		{
			timeSStr << ", ";
		}

		const std::string str = std::to_string(milliseconds);
		timeSStr << std::string(std::max(0, int(3-str.size())), ' ') << str << " ms";

		unitWritten = true;
	}

	if (microseconds > 0u)
	{
		if (unitWritten)
		{
			timeSStr << ", ";
		}

		const std::string str = std::to_string(microseconds);
		timeSStr << std::string(std::max(0, int(3-str.size())), ' ') << str << " us";

		unitWritten = true;
	}

	if (nanoseconds > 0u)
	{
		if (unitWritten)
		{
			timeSStr << ", ";
		}

		const std::string str = std::to_string(nanoseconds);
		timeSStr << std::string(std::max(0, int(3-str.size())), ' ') << str << " ns";

		unitWritten = true;
	}

	return timeSStr.str();
}