#include <chrono>
#include <filesystem>
#include <fstream>
#include <iostream>
#include <memory>

#include <GL/glew.h>
#include <GLFW/glfw3.h>
#include <Eigen/Core>

#include "AbstractTessellation.h"
#include "Camera.h"
#include "CornelEtAl2022/TessellationCornelEtAl2022.h"
#include "Heightfield.h"
#include "KerblEtAl2022/TessellationKerblEtAl2022.h"
#include "KhouryEtAl2019/TessellationKhouryEtAl2019.h"
#include "LeeEtAl2013/TessellationLeeEtAl2013.h"
#include "Naive/TessellationNaive.h"
#include "PipelineStatisticsQuery.h"

// Use this define for benchmarks to reduce the overhead of unrelated operations
//#define BENCHMARK_MODE

//=================================================================================================
/**
*	Simply outputs an error to the console.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
static void writeError(const std::string& description)
{
	std::cout << "Error: " << description;
}

//=================================================================================================
/**
*	Simply outputs an error to the console.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
static void writeError(int error, const char* description)
{
	writeError(std::string(description));
}

//=================================================================================================
/**
*	Formats an OpenGL debug message and logs it.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void formatOpenGLMessage(GLenum source, GLenum type, GLuint id, GLenum severity, const GLchar* message, const void* userParam)
{
	// String for the source
	std::string sourceStr;

	switch (source)
	{
		case GL_DEBUG_CATEGORY_API_ERROR_AMD:
		case GL_DEBUG_SOURCE_API:
		{
			sourceStr = "API";

			break;
		}
		case GL_DEBUG_CATEGORY_WINDOW_SYSTEM_AMD:
		case GL_DEBUG_SOURCE_WINDOW_SYSTEM:
		{
			sourceStr = "Window system";

			break;
		}
		case GL_DEBUG_CATEGORY_SHADER_COMPILER_AMD:
		case GL_DEBUG_SOURCE_SHADER_COMPILER:
		{
			sourceStr = "Shader compiler";

			break;
		}
		case GL_DEBUG_SOURCE_THIRD_PARTY:
		{
			sourceStr = "Third party";

			break;
		}
		case GL_DEBUG_CATEGORY_APPLICATION_AMD:
		case GL_DEBUG_SOURCE_APPLICATION:
		{
			sourceStr = "Application";

			break;
		}
		case GL_DEBUG_CATEGORY_OTHER_AMD:
		case GL_DEBUG_SOURCE_OTHER:
		{
			sourceStr = "Other";

			break;
		}
		default:
		{
			sourceStr = "Unknown";

			break;
		}
	}

	// String for the type
	std::string typeStr;

	switch (type)
	{
		case GL_DEBUG_TYPE_ERROR:
		{
			typeStr = "Error";

			break;
		}
		case GL_DEBUG_CATEGORY_DEPRECATION_AMD:
		case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
		{
			typeStr = "Deprecated";

			break;
		}
		case GL_DEBUG_CATEGORY_UNDEFINED_BEHAVIOR_AMD:
		case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
		{
			typeStr = "Undefined";

			break;
		}
		case GL_DEBUG_TYPE_PORTABILITY:
		{
			typeStr = "Portability";

			break;
		}
		case GL_DEBUG_CATEGORY_PERFORMANCE_AMD:
		case GL_DEBUG_TYPE_PERFORMANCE:
		{
			typeStr = "Performance";

			break;
		}
		case GL_DEBUG_CATEGORY_OTHER_AMD:
		case GL_DEBUG_TYPE_OTHER:
		{
			typeStr = "Other";

			break;
		}
		default:
		{
			typeStr = "Unknown";

			break;
		}
	}

	// String for the severity
	std::string severityStr;

	switch (severity)
	{
		case GL_DEBUG_SEVERITY_HIGH:
		{
			severityStr = "High";

			break;
		}
		case GL_DEBUG_SEVERITY_MEDIUM:
		{
			severityStr = "Medium";

			break;
		}
		case GL_DEBUG_SEVERITY_LOW:
		case GL_DEBUG_SEVERITY_NOTIFICATION:
		{
			return;
		}
		default:
		{
			severityStr = "Unknown";

			break;
		}
	}

	if (id == 131218)
	{
		// Ignore "Program/shader state performance warning: Fragment shader in program being recompiled based on GL state."
		return;
	}
	else if (id == 131186)
	{
		// Buffer access performance warning
		return;
	}

	// String for the detailed message
	std::string messageStr(message);

	// Combined string
	std::stringstream ss;
	ss << "OpenGL debug message" << std::endl;
	ss << " Source: " << sourceStr << std::endl;
	ss << " Type: " << typeStr << std::endl;
	ss << " Severity: " << severityStr << std::endl;
	ss << " ID: " << id << std::endl;
	ss << " Context: " << userParam << std::endl;
	ss << " Message: " << messageStr << std::endl;

	writeError(ss.str());
}

//=================================================================================================
/**
*	Debug callback for AMD cards
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void __stdcall debugCallbackAMD(GLuint id, GLenum category, GLenum severity, GLsizei length, const GLchar* message, void* userParam)
{
	formatOpenGLMessage(category, category, id, severity, message, userParam);
}

//=================================================================================================
/**
*	Debug callback for NVIDIA and Intel cards
*
*	@author	Daniel Cornel
**/
//=================================================================================================
void __stdcall debugCallbackARB(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const GLchar* message, const void* userParam)
{
	formatOpenGLMessage(source, type, id, severity, message, userParam);
}

//=================================================================================================
/**
*	Callback to handle key inputs.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
static void keyCallback(GLFWwindow* window, int key, int scancode, int action, int mods)
{
	if ((key == GLFW_KEY_ESCAPE) && (action == GLFW_PRESS))
	{
		glfwSetWindowShouldClose(window, GLFW_TRUE);
	}
}

const std::vector<std::pair<Eigen::Vector4f, float>> cameraPath
{
	std::make_pair(Eigen::Vector4f(28361.7f, 16386.86f, 30416.1f, 1.0f), 0.0f),
	std::make_pair(Eigen::Vector4f(2939.93f, 896.415f, 670.71f, 1.0f), 0.0f),
	std::make_pair(Eigen::Vector4f(2697.58f, 864.101f, 900.65f, 1.0f), 0.0f),
	std::make_pair(Eigen::Vector4f(2759.91f, -2355.91f, 704.575f, 1.0f), 0.0f),
	std::make_pair(Eigen::Vector4f(-312.336f, -3572.62f, 478.977f, 1.0f), 0.0f),
	std::make_pair(Eigen::Vector4f(-926.15f, -2948.26f, 594.805f, 1.0f), 0.0f),
	std::make_pair(Eigen::Vector4f(-3211.05f, -1235.22f, 663.464f, 1.0f), 0.0f),
	std::make_pair(Eigen::Vector4f(-2584.94f, 4396.94f, 990.321f, 1.0f), 0.0f),
	std::make_pair(Eigen::Vector4f(892.91f, 4557.99f, 853.867f, 1.0f), 0.0f),
	std::make_pair(Eigen::Vector4f(1816.87f, 3470.57f, 474.65f, 1.0f), 0.0f),
	std::make_pair(Eigen::Vector4f(1077.65f, 1556.07f, 919.931f, 1.0f), 0.0f)
};

//=================================================================================================
/**
*	Animates the camera position along a B-spline path over time.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
Eigen::Vector4f getCameraPosition(float timeInMicroseconds)
{
	const float cameraSpeedInMillisecondsPerPercent = 50000.0f;

	int segmentIndex = static_cast<int>(cameraPath.size()) - 1;
	float t = 0.0f;
	float accumulatedDuration = 0.0f;

	for (int controlPointIndex = 0; controlPointIndex < static_cast<int>(cameraPath.size()) - 1; controlPointIndex++)
	{
		const std::pair<Eigen::Vector4f, float>& currentControlPoint = cameraPath[controlPointIndex];
		const std::pair<Eigen::Vector4f, float>& nextControlPoint = cameraPath[controlPointIndex + 1];
		float segmentDuration = 100.0f * cameraSpeedInMillisecondsPerPercent;

		float durationDifference = timeInMicroseconds - accumulatedDuration;

		if (durationDifference <= currentControlPoint.second * 1000000.0f)
		{
			accumulatedDuration += durationDifference;
			segmentIndex = controlPointIndex;
			t = 0.0f;

			break;
		}
		else
		{
			accumulatedDuration += currentControlPoint.second * 1000000.0f;
			durationDifference -= currentControlPoint.second * 1000000.0f;

			if (durationDifference <= segmentDuration)
			{
				t = durationDifference / segmentDuration;

				accumulatedDuration += durationDifference;
				segmentIndex = controlPointIndex;

				break;
			}
			else
			{
				accumulatedDuration += segmentDuration;
			}
		}
	}

	const Eigen::Vector4f& v0 = cameraPath[std::max(0, segmentIndex - 1)].first;
	const Eigen::Vector4f& v1 = cameraPath[segmentIndex].first;
	const Eigen::Vector4f& v2 = cameraPath[std::min(segmentIndex + 1, static_cast<int>(cameraPath.size()) - 1)].first;
	const Eigen::Vector4f& v3 = cameraPath[std::min(segmentIndex + 2, static_cast<int>(cameraPath.size()) - 1)].first;

	return ((1.0 - t) * (1.0 - t) * (1.0 - t) * v0 +
		(3.0 * t * t * t - 6.0 * t * t + 4.0) * v1 +
		(3.0 * (1.0 - t) * (1.0 - t) * (1.0 - t) - 6.0 * (1.0 - t) * (1.0 - t) + 4.0) * v2 +
		t * t * t * v3) / 6.0;
}

//=================================================================================================
/**
*	Main method with render loop.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
int main(int argc, char* argv[])
{
	Eigen::Vector2i windowSize(1920, 1080);
	std::string windowTitle = "Terrain Tessellation";

	//=================================================================================================
	/**
	*	GLFW initialization and window creation
	*/
	//=================================================================================================

	glfwSetErrorCallback(writeError);

	if (!glfwInit())
	{
		exit(EXIT_FAILURE);
	}

#ifdef BENCHMARK_MODE
	// Disable vsync for benchmarking
	glfwWindowHint(GLFW_DOUBLEBUFFER, GL_FALSE);
#endif

	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_COMPAT_PROFILE);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 5);

#ifndef BENCHMARK_MODE
	// Enable for development, disable for benchmarking
	glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GL_TRUE);
#endif

	GLFWwindow* window = glfwCreateWindow(windowSize.x(), windowSize.y(), windowTitle.c_str(), NULL, NULL);

	if (!window)
	{
		glfwTerminate();

		exit(EXIT_FAILURE);
	}

	glfwSetKeyCallback(window, keyCallback);
	glfwMakeContextCurrent(window);

#ifdef BENCHMARK_MODE
	// Disable vsync for benchmarking
	glfwSwapInterval(0);
#else
	glfwSwapInterval(1);
#endif

	//=================================================================================================
	/**
	*	GLEW initialization and compatibility checks
	*/
	//=================================================================================================

	glewExperimental = GL_TRUE;
	glewInit();
	glGetError();

#ifndef BENCHMARK_MODE
	// Enable for development, disable for benchmarking
	if (glewIsSupported("GL_KHR_debug") || glewIsSupported("GL_ARB_debug_output"))
	{
		glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
		glDebugMessageCallback(&debugCallbackARB, nullptr);
	}
	else if (glewIsSupported("GL_AMD_debug_output"))
	{
		glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
		glDebugMessageCallbackAMD(&debugCallbackAMD, nullptr);
	}
#endif

	std::string vendorString(reinterpret_cast<const char*>(glGetString(GL_VENDOR)));
	std::string vendorErrorHint = " If your system contains multiple GPUs, please make sure that it uses the dedicated GPU for this application. Reported GPU vendor: " + vendorString;

	if (!GLEW_ARB_tessellation_shader)
	{
		writeError("Hardware tessellation is not supported on this system!" + vendorErrorHint);
	}

	if (!GL_ARB_compute_shader)
	{
		writeError("Compute shaders are not supported on this system!" + vendorErrorHint);
	}

	if (!GLEW_EXT_direct_state_access)
	{
		writeError("Direct state access is not supported on this system!" + vendorErrorHint);
	}

	try
	{
		//=================================================================================================
		/**
		*	Scene setup
		*/
		//=================================================================================================

		glDepthMask(GL_TRUE);
		glEnable(GL_DEPTH_TEST);
		glDepthFunc(GL_GEQUAL);
		glDepthRange(0.0, 1.0);
		glClipControl(GL_LOWER_LEFT, GL_ZERO_TO_ONE);

		glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
		glClearColor(0.9f, 0.9f, 0.9f, 1.0f);
		glClearDepth(0.0);

		glBindFramebuffer(GL_FRAMEBUFFER, 0);
		glViewport(0, 0, windowSize.x(), windowSize.y());

		Camera camera;

		Eigen::Vector2i heightfieldDimensions(10000, 10000);
		Eigen::Vector2f heightfieldCellSize(1.0f, 1.0f);
		Eigen::Vector2f heightfieldSize = heightfieldDimensions.cast<float>().cwiseProduct(heightfieldCellSize);
		Eigen::Vector2f heightfieldOrigin = -0.5f * heightfieldDimensions.cast<float>().cwiseProduct(heightfieldCellSize);
		float heightfieldScaleFactor = 1000.0f;

		Heightfield heightfield(heightfieldDimensions, heightfieldCellSize, heightfieldOrigin, heightfieldScaleFactor);

		//=================================================================================================
		/**
		*	Create a tessellation with the algorithm of your choice:
		*	1) TessellationNaive: A naive regular tessellation with two triangles per heightfield cell.
		*	2) TessellationLeeEtAl2013: Recursive hardware tessellation proposed by Lee et al.
		*	3) TessellationKhouryEtAl2019:	Adaptive tessellation with compute shaders proposed by Khoury
		*									et al.
		*	4) TessellationKerblEtAl2022:	Improved version of 3) with extensions proposed by Kerbl et
		*									al.
		*	5) TessellationCornelEtAl2022:	Our approach.
		*/
		//=================================================================================================

		//std::unique_ptr<AbstractTessellation> tessellation = std::make_unique<TessellationNaive>(heightfield);
		//std::unique_ptr<AbstractTessellation> tessellation = std::make_unique<TessellationLeeEtAl2013>(heightfield);
		//std::unique_ptr<AbstractTessellation> tessellation = std::make_unique<TessellationKhouryEtAl2019>(heightfield);
		//std::unique_ptr<AbstractTessellation> tessellation = std::make_unique<TessellationKerblEtAl2022>(heightfield);
		std::unique_ptr<AbstractTessellation> tessellation = std::make_unique<TessellationCornelEtAl2022>(heightfield);

		//=================================================================================================
		/**
		*	Setup of the camera path. The camera animation is a bit wonky and behaves weird under some
		*	frame rates, but it is good enough for benchmarking.
		*/
		//=================================================================================================

		long long animationDuration = 0ll;

#ifdef BENCHMARK_MODE
		std::vector<long long> frameDurations;
		frameDurations.reserve(60 * 1000);
#endif

		Eigen::Vector4f forwardCameraPosition = getCameraPosition(1000.0f);
		Eigen::Vector4f previousCameraPosition = cameraPath.front().first - (forwardCameraPosition - cameraPath.front().first);
		Eigen::Vector4f currentCameraPosition = cameraPath.front().first;
		Eigen::Vector4f lastGoodForwardVector = (forwardCameraPosition - cameraPath.front().first).normalized();
		Eigen::Vector4f currentForwardVector = lastGoodForwardVector;

		//=================================================================================================
		/**
		*	Main loop
		*/
		//=================================================================================================

		while (!glfwWindowShouldClose(window))
		{
			auto frameStart = std::chrono::system_clock::now();

			int width;
			int height;
			glfwGetFramebufferSize(window, &width, &height);

			if ((width != windowSize.x()) || (height != windowSize.y()))
			{
				windowSize.x() = width;
				windowSize.y() = height;

				glViewport(0, 0, windowSize.x(), windowSize.y());
			}

			currentCameraPosition = getCameraPosition(static_cast<float>(animationDuration));

			if ((currentCameraPosition - previousCameraPosition).norm() > 1.0f)
			{
				currentForwardVector = (currentCameraPosition - previousCameraPosition).normalized();
				lastGoodForwardVector = currentForwardVector;
			}
			else
			{
				currentForwardVector = lastGoodForwardVector;
			}

			camera.update(
				windowSize.cast<float>(),
				currentCameraPosition,
				Eigen::Vector4f(currentForwardVector.x(), currentForwardVector.y(), currentForwardVector.z() - 0.3f, 0.0f).normalized(),
				Eigen::Vector4f(0.0f, 0.0f, 1.0f, 0.0f),
				45.0f,
				static_cast<float>(windowSize.x()) / static_cast<float>(windowSize.y()),
				0.1f,
				100000.0f);

			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

			tessellation->update(camera);
			tessellation->render(camera);

			previousCameraPosition = currentCameraPosition;

#ifdef BENCHMARK_MODE
			glFlush();
#else
			glfwSwapBuffers(window);
#endif

			long long frameDuration = std::chrono::duration_cast<std::chrono::microseconds>(std::chrono::system_clock::now() - frameStart).count();

			animationDuration += frameDuration;

#ifdef BENCHMARK_MODE
			frameDurations.push_back(frameDuration);
#endif

			if (animationDuration > 60000000)
			{
				break;
			}

			glfwPollEvents();
		}

#ifdef BENCHMARK_MODE
		//=================================================================================================
		/**
		*	Aggregate the minimum frame durations in bins of 100 ms each
		*/
		//=================================================================================================

		long long timeStepSize = 100000ll;
		std::vector<double> frameDurationPerTimeStep(60000000 / timeStepSize, std::numeric_limits<double>::max());
		long long accumulatedFrameDuration = 0ll;

		for (long long frameDuration : frameDurations)
		{
			double frameDurationInMs = static_cast<double>(frameDuration) / 1000.0;

			for (long long currentTimeStep = accumulatedFrameDuration; currentTimeStep < accumulatedFrameDuration + frameDuration; currentTimeStep++)
			{
				if (currentTimeStep >= 60000000)
				{
					break;
				}

				frameDurationPerTimeStep[currentTimeStep / timeStepSize] = std::min(frameDurationPerTimeStep[currentTimeStep / timeStepSize], frameDurationInMs);
			}

			accumulatedFrameDuration += frameDuration;
		}

		//=================================================================================================
		/**
		*	Store the frame durations in a CSV file
		*/
		//=================================================================================================

		std::string executablePathString = argv[0];
		std::filesystem::path executablePath(executablePathString);
		std::filesystem::path currentDirectory = executablePath.remove_filename();

		std::time_t currentTime = std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());

		std::stringstream csvFileName;
		csvFileName << "frame_times_";
		csvFileName << tessellation->mName;
		csvFileName << "_";
		csvFileName << currentTime;
		csvFileName << ".csv";

		std::filesystem::path csvFilePath = currentDirectory / csvFileName.str();

		std::ofstream csvFile;
		csvFile.open(csvFilePath);

		for (size_t bin = 0; bin < frameDurationPerTimeStep.size(); bin++)
		{
			csvFile << frameDurationPerTimeStep[bin] << "\n";
		}

		csvFile.close();
#endif
	}
	catch (const std::exception& ex)
	{
		writeError(ex.what());
	}

	glfwDestroyWindow(window);
	glfwTerminate();

	exit(EXIT_SUCCESS);
}